
import json

from epyk.core.js import JsUtils
from epyk.core.js.packages import JsPackage

# All the predefined variable types
from epyk.core.js.fncs import JsFncs


def decorate_var(var_name, convert_var=True):
  """
  Return the String Jquery variable reference for a given component id

  :param var_name: String. The variable name
  :param convert_var: Boolean.

  :return: The decorated Jquery reference
  """
  if not convert_var:
    return '$(%s)' % var_name

  return '$(%s)' % JsUtils.jsConvertData(var_name, None)


class Jsjqxhr(object):
  def __init__(self, ajax):
    self.__ajax = {'request': ajax}

  def done(self, jsFncs):
    """
    AJAX Function

    An alternative construct to the success callback option, refer to deferred.done() for implementation details

    Related Pages:

			https://api.jquery.com/jQuery.ajax/#jqXHR

    :param jsFncs: The Javascript Functions

    :return:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__ajax.setdefault("done", []).extend(jsFncs)
    return self

  def fail(self, jsFncs):
    """
    AJAX Function

    An alternative construct to the error callback option, the .fail() method replaces the deprecated .error() method

    Related Pages:

			https://api.jquery.com/jQuery.ajax/#jqXHR

    :param jsFncs: The Javascript Functions

    :return:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__ajax.setdefault("fail", []).extend(jsFncs)
    return self

  def always(self, jsFncs):
    """
    AJAX Function

    An alternative construct to the complete callback option

    Related Pages:

			https://api.jquery.com/jQuery.ajax/#jqXHR

    :param jsFncs: The Javascript Functions

    :return:
    """
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.__ajax.setdefault("always", []).extend(jsFncs)
    return self

  def toStr(self):
    """
    Javascript representation

    :return: Return the Javascript String
    """
    reqResult = [self.__ajax['request']]
    for rType in ["done", "fail", "always"]:
      if self.__ajax.get(rType) is not None:
        reqResult.append("%s(function(){%s})" % (rType, ";".join(JsUtils.jsConvertFncs(self.__ajax[rType]))))
    return ".".join(map(lambda x: str(x), reqResult))


class JQuery(JsPackage):
  """
  Jquery wrapper.
  for more details about the different available functions go on the website: https://jquery.com/
  In order to avoid conflict with any other libraries the $ will not be used in this framework.
  Any JQuery reference will be done using the proper name of the library jQuery()

  Documentation:
    - https://www.w3schools.com/jquery/
    - https://www.w3schools.com/jquery/jquery_noconflict.asp
    - https://www.w3schools.com/jquery/jquery_ref_ajax.asp

  """
  lib_alias = {"js": 'jquery'}
  lib_selector = 'jQuery("body")'

  def this(self, reference=None):
    """
    Force the selector to be this or a specific reference.
    This feature can be useful in functions

    :param reference: The Jquery reference to be used instead example #MyId

    :return: The Jquery Python object
    """
    if len(self._js) > 0:
      raise Exception("Selector can only be changed first")

    if reference is None:
      self._selector = "jQuery(this)"
    else:
      self._selector = "jQuery('%s')" % reference
    return self

  def new(self, tag=None, reference=None):
    if len(self._js) > 0:
      raise Exception("Selector can only be changed first")

    if tag is None and reference is None:
      raise Exception("Tag or / and Reference must be defined")

    if tag is None and reference is not None:
      self._selector = "jQuery('%s')" % reference
    else:
      if reference is not None:
        self._selector = "jQuery('<%s id=\\'%s\\'></%s>')" % (tag, reference, tag)
      else:
        self._selector = "jQuery('<%s></%s>')" % (tag, tag)
    return self

  def toggle(self, speed=None, easing=None, jsCallback=None):
    """

    :param speed:
    :param easing:
    :param jsCallback:

    :return:
    """
    return self.fnc("toggle()")

  def trigger(self, jsData, jsFnc=None):
    """

    :param jsData:
    :param jsFnc:

    :return:
    """
    jsData = JsUtils.jsConvertData(jsData, jsFnc)
    return self.fnc("trigger(%(jsData)s)" % {"jsData": jsData})

  def hide(self, speed=None, callback=None):
    """

    Related Pages:

			https//www.w3schools.com/jquery/jquery_hide_show.asp
    http://api.jquery.com/hide/

    :param speed:
    :param callback:

    :return:
    """
    if speed is not None:
      if callback is not None:
        if not isinstance(callback, list):
          callback = [callback]
        jqFnc = "hide(%(speed)s, function(){%(callback)s})" % {'speed': speed,  'callback': ";".join(callback)}
      else:
        jqFnc = "hide(%(speed)s)" % {'speed': speed}
    else:
      jqFnc = "hide()"
    return self.fnc(jqFnc)

  def show(self, speed=None, callback=None):
    """

    Related Pages:

			https//www.w3schools.com/jquery/jquery_hide_show.asp
    http://api.jquery.com/show/

    :param speed:
    :param callback:

    :return:
    """
    if speed is not None:
      if callback is not None:
        if not isinstance(callback, list):
          callback = [callback]
        jqFnc = "show(%(speed)s, function(){%(callback)s})" % {'speed': speed, 'callback': ";".join(callback)}
      else:
        jqFnc = "show(%(speed)s)" % {'speed': speed}
    else:
      jqFnc = "show()"
    return self.fnc(jqFnc)

  def fadeIn(self, speed=None, callback=None):
    """

    Related Pages:

			https//www.w3schools.com/jquery/jquery_fade.asp
    http://api.jquery.com/fadein/

    :param speed:
    :param callback:

    :return:
    """
    if speed is not None:
      if callback is not None:
        if not isinstance(callback, list):
          callback = [callback]
        jqFnc = "fadeIn(%(speed)s, function(){%(callback)s})" % {'speed': speed, 'callback': ";".join(callback)}
      else:
        jqFnc = "fadeIn(%(speed)s)" % {'speed': speed}
    else:
      jqFnc = "fadeIn()"
    return self.fnc(jqFnc)

  def fadeOut(self, speed=None, callback=None):
    """

    Related Pages:

			https//www.w3schools.com/jquery/jquery_fade.asp
    http://api.jquery.com/fadeout/

    :param speed:
    :param callback:
    :return:
    """
    if speed is not None:
      if callback is not None:
        if not isinstance(callback, list):
          callback = [callback]
        jqFnc = "fadeOut(%(speed)s, function(){%(callback)s})" % {'speed': speed, 'callback': ";".join(callback)}
      else:
        jqFnc = "fadeOut(%(speed)s)" % {'speed': speed}
    else:
      jqFnc = "fadeOut()"
    return self.fnc(jqFnc)

  def fadeToggle(self, speed=None, callback=None):
    """

    Documentation:
      - https://www.w3schools.com/jquery/jquery_fade.asp

    :param speed:
    :param callback:

    :return:
    """
    if speed is not None:
      if callback is not None:
        if not isinstance(callback, list):
          callback = [callback]
        jqFnc = "fadeToggle(%(speed)s, function(){%(callback)s})" % {'speed': speed, 'callback': ";".join(callback)}
      else:
        jqFnc = "fadeToggle(%(speed)s)" % {'speed': speed}
    else:
      jqFnc = "fadeToggle()"
    return self.fnc(jqFnc)

  def fadeTo(self, duration, opacity, easing=None, complete=None):
    """
    Adjust the opacity of the matched elements.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_fade.asp
    https://api.jquery.com/fadeto/

    :param duration: A string or number determining how long the animation will run.
    :param opacity: A number between 0 and 1 denoting the target opacity.
    :param easing: A string indicating which easing function to use for the transition.
    :param complete: A function to call once the animation is complete.
    """
    if complete is not None:
      if not isinstance(complete, list):
        complete = [complete]
      complete = "function(){%s}" % ";".join(complete)
    if easing is not None:
      if complete is not None:
        jqFnc = "fadeTo(%(speed)s, %(opacity)s, %(easing)s, %(callback)s)" % {'speed': duration, 'opacity': opacity, 'easing': easing, 'callback': complete}
      else:
        jqFnc = "fadeTo(%(speed)s, %(opacity)s, %(easing)s)" % {'speed': duration, 'opacity': opacity, 'easing': easing}
    elif complete is not None:
      jqFnc = "fadeTo(%(speed)s, %(opacity)s, %(complete)s)" % {'speed': duration, 'opacity': opacity, 'complete': complete}
    else:
      jqFnc = "fadeTo(%(speed)s, %(opacity)s)" % {'speed': duration, 'opacity': opacity}
    return self.fnc(jqFnc)

  def slideDown(self, speed=None, callback=None):
    """
    Display the matched elements with a sliding motion.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_slide.asp
    https://api.jquery.com/slideDown/#slideDown-duration-complete

    :param speed: A string or number determining how long the animation will run.
    :param callback: A string indicating which easing function to use for the transition.
    """
    if speed is not None:
      if callback is not None:
        if not isinstance(callback, list):
          callback = [callback]
        jqFnc = "slideDown(%(speed)s, function(){%(callback)s})" % {'speed': speed, 'callback': ";".join(callback)}
      else:
        jqFnc = "slideDown(%(speed)s)" % {'speed': speed}
    else:
      jqFnc = "slideDown()"
    return self.fnc(jqFnc)

  def slideUp(self, speed=None, callback=None):
    """
    Hide the matched elements with a sliding motion.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_slide.asp
    https://api.jquery.com/slideUp/#slideUp-duration-complete

    :param speed: A string or number determining how long the animation will run.
    :param callback: A function to call once the animation is complete, called once per matched element
    """
    if speed is not None:
      if callback is not None:
        if not isinstance(callback, list):
          callback = [callback]
        jqFnc = "slideUp(%(speed)s, function(){%(callback)s})" % {'speed': speed, 'callback': ";".join(callback)}
      else:
        jqFnc = "slideUp(%(speed)s)" % {'speed': speed}
    else:
      jqFnc = "slideUp()"
    return self.fnc(jqFnc)

  def slideToggle(self, speed=None, callback=None):
    """
    Display or hide the matched elements with a sliding motion.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_slide.asp
    https://api.jquery.com/slideToggle/#slideToggle-duration-complete

    :param speed: A string or number determining how long the animation will run.
    :param callback: A function to call once the animation is complete, called once per matched element.
    """
    if speed is not None:
      if callback is not None:
        if not isinstance(callback, list):
          callback = [callback]
        jqFnc = "slideToggle(%(speed)s, function(){%(callback)s})" % {'speed': speed, 'callback': ";".join(callback)}
      else:
        jqFnc = "slideToggle(%(speed)s)" % {'speed': speed}
    else:
      jqFnc = "slideToggle()"
    return self.fnc(jqFnc)

  def animate(self, params, speed=400, easing='swing', callback=None):
    """
    Perform a custom animation of a set of CSS properties.

    Example
    myObj.animate(0.25, "+=50")

    Related Pages:

			https://www.w3schools.com/jquery/jquery_animate.asp
    https://api.jquery.com/animate/#animate-properties-duration-easing-complete

    :param params: An object of CSS properties and values that the animation will move toward.
    :param speed: A string or number determining how long the animation will run.
    :param easing: A string indicating which easing function to use for the transition.
    :param callback: A function to call once the animation is complete, called once per matched element.
    """
    easing = JsUtils.jsConvertData(easing, None)
    if callback is not None:
      return self.fnc("animate(%s, %s, %s, %s)" % (params, speed, easing, callback))

    return self.fnc("animate(%s, %s, %s)" % (params, speed, easing))

  def stop(self, stopAll=False, goToEnd=False):
    """
    Stop the currently-running animation on the matched elements.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_stop.asp
    https://api.jquery.com/stop/#stop-clearQueue-jumpToEnd

    :param stopAll: A Boolean indicating whether to remove queued animation as well
    :param goToEnd: A Boolean indicating whether to complete the current animation immediately
    """
    stopAll = JsUtils.jsConvertData(stopAll, None)
    goToEnd = JsUtils.jsConvertData(goToEnd, None)
    return self.fnc("stop(%(stopAll)s, %(goToEnd)s)" % {'stopAll': stopAll, 'goToEnd': goToEnd})

  def remove(self, selector=None):
    """
    Remove the set of matched elements from the DOM.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_dom_remove.asp
    https://api.jquery.com/remove/#remove-selector

    :param selector: A selector expression that filters the set of matched elements to be removed.
    """
    if selector is not None:
      selector = JsUtils.jsConvertData(selector, None)
      return self.fnc("remove(%s)" % selector)

    return self.fnc("remove()")

  def empty(self):
    """
    Remove all child nodes of the set of matched elements from the DOM.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_dom_remove.asp
    https://api.jquery.com/empty/#empty

    :return:
    """
    return self.fnc("empty()")

  def siblings(self, selector=None):
    """
    Get the siblings of each element in the set of matched elements, optionally filtered by a selector.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_traversing_siblings.asp
    https://api.jquery.com/siblings/#siblings-selector

    :param selector: A string containing a selector expression to match elements against.
    """
    if selector is not None:
      selector = JsUtils.jsConvertData(selector, None)
      return self.fnc("siblings(%s)" % selector)

    return self.fnc("siblings()")

  def next(self, selector=None):
    """
    Get the immediately following sibling of each element in the set of matched elements.
    If a selector is provided, it retrieves the next sibling only if it matches that selector.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_traversing_siblings.asp
    https://api.jquery.com/next/#next-selector

    :param selector: A string containing a selector expression to match elements against.
    :return:
    """
    if selector is not None:
      selector = JsUtils.jsConvertData(selector, None)
      return self.fnc("next(%s)" % selector)

    return self.fnc("next()")

  def prev(self, selector=None):
    """
    Get the immediately preceding sibling of each element in the set of matched elements.
    If a selector is provided, it retrieves the previous sibling only if it matches that selector.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_traversing_siblings.asp
    https://api.jquery.com/prev/#prev-selector

    :param selector: A string containing a selector expression to match elements against.
    :return:
    """
    if selector is not None:
      selector = JsUtils.jsConvertData(selector, None)
      return self.fnc("prev(%s)" % selector)

    return self.fnc("prev()")

  def first(self):
    """
    The first() method returns the first element of the specified elements.

    Related Pages:

			https//www.w3schools.com/jquery/jquery_traversing_filtering.asp
    https://api.jquery.com/first/#first

    :return:
    """
    return self.fnc("first()")

  def children(self, selector=None):
    """
    The children() method returns all direct children of the selected element

    Related Pages:

			https//www.w3schools.com/jquery/traversing_children.asp

    :param selector: Optional. Specifies a selector expression to narrow down the search for children
    :return:
    """
    if selector is None:
      return self.fnc("children()")

    return self.fnc("children(%s)" % selector)

  def last(self):
    """
    The last() method returns the last element of the specified element

    Related Pages:

			https//www.w3schools.com/jquery/jquery_traversing_filtering.asp

    :return:
    """
    return self.fnc("last()")

  def appendTo(self, dstJqId, jsFnc=None):
    """
    Insert every element in the set of matched elements to the end of the target.

    :param dstJqId: A selector, element, HTML string, array of elements, or jQuery object; the matched set of elements will be inserted at the end of the element(s) specified by this parameter.
    :param jsFnc:

    :rtype: str

    :return:
    """
    return self.fnc("appendTo(%(dstJqId)s)" % {'dstJqId': JsUtils.jsConvertData(dstJqId, jsFnc)})

  def append(self, dstJqId, jsFnc=None):
    """

    :param dstJqId:
    :param jsFnc:
    :rtype: str

    :return:
    """
    return self.fnc("append(%(dstJqId)s)" % {'dstJqId': JsUtils.jsConvertData(dstJqId, jsFnc)})

  def prepend(self, jsData, jsFnc=None):
    """
    Insert content, specified by the parameter, to the beginning of each element in the set of matched elements.

    Related Pages:

			https://api.jquery.com/prepend/#prepend-content-content

    :param jsData:
    :param jsFnc:

    :return:
    """
    return self.fnc("prepend(%(data)s)" % {"data": JsUtils.jsConvertData(jsData, jsFnc)})

  def eq(self, i):
    """
    The eq() method returns an element with a specific index number of the selected elements.

    Documentation:
      - https://www.w3schools.com/jquery/jquery_traversing_filtering.asp

    :param i: The index numbers start at 0, so the first element will have the index number 0 and not 1
    :return:
    """
    return self.fnc("eq(%(index)s)" % {'index': i})

  def filter(self, selector=None):
    """

    :return:
    """

  def _not(self):
    """

    :return:
    """

  def find(self, criteria):
    """
    Get the descendants of each element in the current set of matched elements, filtered by a selector, jQuery object, or element.

    Related Pages:

			https://api.jquery.com/find/#entry-examples

    :param criteria: Selector or element An element or a jQuery object to match elements against.

    :return:
    """
    criteria = JsUtils.jsConvertData(criteria, None)
    return self.fnc("find(%s)" % criteria)

  def each(self, jsFncs):
    """
    Iterate over a jQuery object, executing a function for each matched element.

    Related Pages:

			https://api.jquery.com/each/#each-function

    :param jsFncs: A function to execute for each matched element.

    :return:
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs)
    return self.fnc("each(function(index, data){%s})" % ";".join(jsFncs))

  def css(self, key, value=None):
    """
    Hook directly into jQuery to override how particular CSS properties are retrieved or set, normalize CSS property naming, or create custom properties.

    Related Pages:

			https://api.jquery.com/jQuery.cssHooks/#jQuery-cssHooks1

    :param key:
    :param value:

    :return:
    """
    if hasattr(self.src, "style"):
      self.src.style.css(key, value)
      return self

    else:
      if value is None:
        return self.fnc("css(%s)" % JsUtils.jsConvertData(key, None))

      return self.fnc("css('%s', %s)" % (key, JsUtils.jsConvertData(value, None)))

  def attr(self, key, value):
    """
    Get the value of an attribute for the first element in the set of matched elements.

    Related Pages:

			https://api.jquery.com/attr/#attr-attributeName

    :param key: The name of the attribute to get.
    :param value: A value to set for the attribute. If null, the specified attribute will be removed

    :return:
    """
    if key.lower() in ["style", 'class']:
      raise Exception("Only the css() function can be used to change the style")

    self.src._attrs[key] = value
    return self.src

  def val(self, jsData=None, jsFnc=None):
    """
    Get the current value of the first element in the set of matched elements.

    Related Pages:

			https://api.jquery.com/val/#val

    :param jsData:
    :param jsFnc:
    :return:
    """
    if jsData is None:
      return self.fnc("val()")

    return self.fnc("val(%s)" % JsUtils.jsConvertData(jsData, jsFnc))

  def text(self, jsData, jsFnc=None):
    """
    Get the combined text contents of each element in the set of matched elements, including their descendants, or set the text contents of the matched elements.

    Related Pages:

			https://api.jquery.com/text/#text

    :param jsData:
    :param jsFnc:

    :return:
    """
    if jsData is None:
      return self.fnc("text()")

    jsData = JsUtils.jsConvertData(jsData, jsFnc)
    return self.fnc("text(%s)" % jsData)

  def html(self, jsData=None, jsFnc=None):
    """

    :param jsData:
    :param jsFnc:

    :return:
    """
    if jsData is None:
      return self.fnc("html()")

    return self.fnc("html(%s)" % JsUtils.jsConvertData(jsData, jsFnc))

  def toggleClass(self, clsName, propagate=False):
    """

    :rtype: str
    :return:
    """
    if propagate:
      self.fnc_closure('parentNode.childNodes.forEach(function(e){e.classList.remove("%(data)s")})' % {'data': clsName})
    return self.fnc('toggleClass("%(data)s")' % {'data': clsName})

  def addClass(self, clsName, attrs=None, eventAttrs=None):
    """
    Adds the specified class(es) to each element in the set of matched elements.

    This function can either use an existing class or create one if the attrs or eventAttrs are defined

    :param clsName: The Css classname
    :param attrs: A python dictionary with the css attributes
    :param eventAttrs: A nested python dictionary with the css attributes for each events

    :return:
    """
    if attrs is not None or eventAttrs is not None:
      clsName = self.src.style.cssName(clsName)
      self.src.style.cssCls(clsName, attrs, eventAttrs, False)
    return self.fnc('addClass("%s")' % clsName)

  def getJSON(self, url, jsData, success, dataType='json', jsDataKey=None, isPyData=True, jsFnc=None):
    """
    Load JSON-encoded data from the server using a GET HTTP request.

    Related Pages:

			https//api.jquery.com/jQuery.getJSON/#jQuery-getJSON-url-data-success

    :return:
    """
    success = JsUtils.jsConvertFncs(success)
    jsData = JsUtils.jsConvert(jsData, jsDataKey, isPyData, jsFnc)
    return Jsjqxhr("jQuery.getJSON('%s', {data: JSON.stringify(%s)}, function(data) {%s}, '%s')" % (url, jsData, ";".join(success), dataType))

  def getJsScript(self, url, jsData, success, dataType='json', jsDataKey=None, isPyData=True, jsFnc=None):
    """
    Load a JavaScript file from the server using a GET HTTP request, then execute it.

    Related Pages:

			https//api.jquery.com/jQuery.getScript/

    :param url:
    :param jsData:
    :param success:
    :param dataType:
    :param jsDataKey:
    :param isPyData:
    :param jsFnc:
    :return:
    """
    success = JsUtils.jsConvertFncs(success)
    jsData = JsUtils.jsConvert(jsData, jsDataKey, isPyData, jsFnc)
    return Jsjqxhr("jQuery.getScript('%s', {data: JSON.stringify(%s)}, function(data, textStatus, jqxhr) {%s}, '%s')" % (url, jsData, ";".join(success), dataType))

  def getPyScript(self, script, data=None, successFncs=None, options=None, timeout=None, props=None):
    """

    Related Pages:

			https//api.jquery.com/jQuery.getJSON/#jQuery-getJSON-url-data-success

    :param script:
    :param jsData:
    :param success:
    :param dataType:
    :param jsDataKey:
    :param isPyData:
    :param jsFnc:

    :return:
    """

    if not hasattr(self.src, "aresObj"):
      if not hasattr(self.src, "run"):
        raise Exception("Cannot work without a proper rptObj")

      else:
        rptObj = self.src
    else:
      rptObj = self.src._report
    if data is None:
      data = {}
    qParams = self.getParams('%s/data/%s/%s' % (rptObj._urlsApp['report'], rptObj.run.report_name, script.replace(".py", "")), data, successFncs, None, options, timeout, props)
    return Jsjqxhr("jQuery.post(%s)" % qParams)

  def load(self, url, jsData=None, successFncs=None):
    """
    Load data from the server and place the returned HTML into the matched elements.

    Example
    div.dom.jquery.load(r"./report_list.html")

    Related Pages:

			https://api.jquery.com/load/#load-url-data-complete

    :param url: A string containing the URL to which the request is sent.
    :param jsData: A plain object or string that is sent to the server with the request.
    :param successFncs: A callback function that is executed when the request completes.
    :return:
    """
    if successFncs is None:
      if jsData is None:
        return "%s.load('%s')" % (self.varId, url)
      return "%s.load('%s', {data: JSON.stringify(%s)})" % (self.varId, url, jsData)

    if jsData is None:
      return "%s.load('%s', function(data) {%s})" % (self.varId, url, successFncs)
    return "%s.load('%s', {data: JSON.stringify(%s)}, function(data) {%s})" % (self.varId, url, jsData, successFncs)

  def ajaxError(self, jsFncs):
    """

    Related Pages:

			https://api.jquery.com/ajaxError/

    :param jsFncs:
    :return:
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs)
    return "jQuery(document).ajaxError(function(event, jqxhr, settings, thrownError) {%s})" % ";".join(jsFncs)

  def ajaxStart(self, jsFncs):
    """
    Register a handler to be called when the first Ajax request begins. This is an Ajax Event.

    Related Pages:

			https://api.jquery.com/ajaxStart/

    :return:
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs)
    return "jQuery(document).ajaxStart(function() {%s})" % ";".join(jsFncs)

  def ajaxStop(self, jsFncs):
    """
    Register a handler to be called when all Ajax requests have completed. This is an Ajax Event.

    Related Pages:

			https://api.jquery.com/ajaxStop/

    :param jsFncs:
    :return:
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs)
    return "jQuery(document).ajaxStop(function() {%s})" % ";".join(jsFncs)

  def ajaxSuccess(self, jsFncs):
    """
    Attach a function to be executed whenever an Ajax request completes successfully. This is an Ajax Event.

    Related Pages:

			https://api.jquery.com/ajaxSuccess/

    :param jsFncs:
    :return:
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs)
    return "jQuery(document).ajaxSuccess(function(event, xhr, settings) {%s})" % ";".join(jsFncs)

  def ajaxSend(self, jsFncs):
    """

    Related Pages:

			https://api.jquery.com/ajaxSend/

    :param jsFncs:
    :return:
    """
    jsFncs = JsUtils.jsConvertFncs(jsFncs)
    return "jQuery(document).ajaxSend(function(event, jqxhr, settings) {%s})" % ";".join(jsFncs)

  def ajaxComplete(self, jsFncs):
    jsFncs = JsUtils.jsConvertFncs(jsFncs)
    return "jQuery(document).ajaxComplete(function() {%s})" % ";".join(jsFncs)

  def getParams(self, url, data, successFncs, errorFncs, options, timeout, props):
    """

    :return:
    """
    ajaxData = []
    if props is not None:
      data["_system"] = props
    if options is not None:
      for k, v in options.items():
        ajaxData.append("%s: %s" % (k, json.dumps(v)))
    ajaxData.extend(["data: {data: JSON.stringify(%s)}" % data, "url: '%s'" % url])
    if timeout is not None:
      ajaxData.append("timeout: %s" % timeout)
    if successFncs is not None:
      ajaxData.append("success: function(result,status,xhr){%s}" % ";".join(JsUtils.jsConvertFncs(successFncs)))
    if errorFncs is not None:
      ajaxData.append("error: function(xhr, status, error){%s}" % ";".join(JsUtils.jsConvertFncs(errorFncs)))
    return "{%s}" % ", ".join(ajaxData)

  def get(self, url, data, successFncs=None, options=None, timeout=None, props=None):
    """
    Load data from the server using a HTTP GET request.

    Documentation:
      - https://www.w3schools.com/jquery/jquery_ajax_get_post.asp

    :return:
    """
    return Jsjqxhr("jQuery.get(%s)" % self.getParams(url, data, successFncs, None, options, timeout, props))

  def post(self, url, data=None, successFncs=None, options=None, timeout=None, props=None):
    """
    Load data from the server using a HTTP POST request.

    Documentation:
      - https://www.w3schools.com/jquery/jquery_ajax_get_post.asp

    :rtype: Jsjqxhr
    :return:
    """
    data = data or {}
    return Jsjqxhr("jQuery.post(%s)" % self.getParams(url, data, successFncs, None, options, timeout, props))

  def ajax(self, type, url, data=None, successFncs=None, errorFncs=None, options=None, timeout=None, props=None):
    """
    The ajax() method is used to perform an AJAX (asynchronous HTTP) request.

    Example


    Related Pages:

			https//www.w3schools.com/jquery/ajax_ajax.asp

    :return:
    :param type: Specifies the type of request. (GET or POST)
    :param url: Specifies the URL to send the request to. Default is the current page
    :param data: Specifies data to be sent to the server
    :param successFncs: A function to be run when the request succeeds
    :param errorFncs: A function to run if the request fails.
    :param options: The other parameters specifies one or more name/value pairs for the AJAX request
    :param timeout: The local timeout (in milliseconds) for the request
    :return:
    """
    if type.upper() not in ['POST', 'GET']:
      raise Exception("Method %s not recognised" % url)

    data = data or {}
    return Jsjqxhr("jQuery.ajax(%s)" % self.getParams(url, data, successFncs, errorFncs, options, timeout, props))

  def click(self, jsFnc):
    """

    :param jsFnc:

    :return:
    """
    self.css("cursor", "pointer")
    return self.fnc("click(function(){%s})" % JsUtils.jsConvertFncs(jsFnc, toStr=True))



from epyk.core.css.styles.classes import CssStyle


class CssWebkitScrollbar(CssStyle.Style):
  classname = "::-webkit-scrollbar"
  _attrs = {'width': '10px'}


class CssWebkitScrollbarTrack(CssStyle.Style):
  classname = "::-webkit-scrollbar-track"
  _attrs = {'border-radius': '10px'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.colors[0]})


class CssWebkitScrollbarThumb(CssStyle.Style):
  classname = "::-webkit-scrollbar-thumb"
  _attrs = {'border-radius': '10px'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.colors[-2]})




"""
CSS Module in charge of defining the global classes
"""

from epyk.core.css.styles.classes import CssStyle


class Overflow(object):
  def __init__(self, cssObj):
    self._cssObj = cssObj
    self._cssObj.cssStyles["::-webkit-scrollbar"] = "{ height: 10px; }"
    self._cssObj.cssStyles["::-webkit-scrollbar-track"] = "{ border-radius: 10px; -webkit-box-shadow: inset 0 0 2px rgba(0,0,0,0.5); }"

  def button(self):
    """
    Add the arrows to the scroll bar

    :return:
    """
    self._cssObj.cssStyles["::-webkit-scrollbar-button:single-button"] = "{ background-color: #bbbbbb; display: block; border-style: solid; height: 13px; width: 16px; }"
    self._cssObj.cssStyles["::-webkit-scrollbar-button:single-button:vertical:decrement"] = "{ border-width: 0 8px 8px 8px; border-color: transparent transparent #555555 transparent; }"
    self._cssObj.cssStyles["::-webkit-scrollbar-button:single-button:vertical:increment"] = "{ border-width: 8px 8px 0 8px; border-color: #555555 transparent transparent transparent; }"

  def thumb(self, style=None, css_hover=None, css_active=None):
    """

    :param style:
    :param css_hover:
    :param css_active:

    :return:
    """
    if style is not None:
      self._cssObj.cssStyles["::-webkit-scrollbar-thumb"] = CssStyle.CssCls.toCss(style)
    if css_hover is not None:
      self._cssObj.cssStyles["::-webkit-scrollbar-thumb:hover"] = CssStyle.CssCls.toCss(css_hover)
    if css_active is not None:
      self._cssObj.cssStyles["::-webkit-scrollbar-thumb:active"] = CssStyle.CssCls.toCss(css_active)

  def horizontal(self):
    """

    :return:
    """

  def vertical(self):
    """

    :return:
    """


class CssGlobal(object):

  def __init__(self, cssObj):
    self._cssObj = cssObj

  @property
  def overflow(self):
    return Overflow(self._cssObj)

  def style(self, name, attrs):
    """
    Add a new CSS Class in the page header

    Related Pages:

			https://www.w3schools.com/html/html_css.asp

    :param name: The CSS Style reference as a string
    :param attrs: The CSS attributes as a dictionary

    :return: The CSS Object
    """
    self._cssObj.cssStyles[name] = CssStyle.CssCls.toCss(attrs)
    return self

  def style_per_id(self, html_id, attrs):
    """
    Add a new CSS Class in the page header

    Related Pages:

			https://www.w3schools.com/html/html_css.asp

    :param html_id: The HTML object ID
    :param attrs: The CSS attributes as a dictionary

    :return:
    """
    self._cssObj.cssStyles["#%s" % html_id] = CssStyle.CssCls.toCss(attrs)
    return self

  def style_per_tag(self, html_tag, attrs):
    """
    Add a new CSS Class in the page header

    Related Pages:

			https://www.w3schools.com/html/html_css.asp

    :param html_tag: The HTML object tag
    :param attrs: The CSS attributes as a dictionary

    :return:
    """
    self._cssObj.cssStyles[html_tag] = CssStyle.CssCls.toCss(attrs)
    return self

  def new_class(self, clss_nam, attrs):
    """
    Add a new CSS Class in the page header

    Related Pages:

			https://www.w3schools.com/html/html_css.asp

    :param clss_nam: The CSS class name as a string
    :param attrs: The CSS attributes as a dictionary

    :return:
    """
    self._cssObj.cssStyles[".%s" % clss_nam] = CssStyle.CssCls.toCss(attrs)
    return self



from epyk.core.js import Js
from epyk.core.js import JsUtils

from epyk.core.js.primitives import JsDate
from epyk.core.js.primitives import JsString


# ------------------------------------------------------------------------------------------------------------------
#
jsObj = Js.JsBase()
f = JsUtils.JsFile("Array", path=r"../outs")
f.writeJs([
  jsObj.objects.array.new([2, 5, 12, -3], "MyArray"),
  jsObj.objects.array.new([3, -9, 2, -6], "MyArray2"),
  jsObj.objects.array.new([], "MyArray3"),
  jsObj.console.log(jsObj.objects.array.get("MyArray3").concat(jsObj.objects.array.get("MyArray"), jsObj.objects.array.get("MyArray2"))),
  jsObj.console.log(jsObj.objects.array.get("MyArray").findIndex([
    jsObj.if_(jsObj.data.loop.val <= 0, [
      jsObj.return_(jsObj.objects.true)]).else_(jsObj.return_(jsObj.objects.false)),
  ])),

  jsObj.console.log(jsObj.objects.array.get("MyArray").map([
    jsObj.data.loop.val * jsObj.math.max(jsObj.data.loop.arr.toArgs()),
    jsObj.return_(jsObj.data.loop.val)
  ])),
  jsObj.console.log("########"),
  # jsObj.console.log(jsObj.objects.array.get("MyArray")),
  # jsObj.console.log(jsObj.objects.array.get("MyArray").reverse()),
  # jsObj.console.log(jsObj.objects.array.get("MyArray").push(55, -17)),
  # jsObj.console.log(jsObj.objects.array.get("MyArray")),
  # jsObj.console.log(jsObj.objects.array.get("MyArray").pop()),
  # jsObj.console.log(jsObj.objects.array.get("MyArray").join("_")),
  # jsObj.console.log(jsObj.objects.array.get("MyArray").slice(3, 5)),
  # jsObj.console.log(jsObj.objects.array.get("MyArray").sort()),
  jsObj.console.log(jsObj.objects.array.new([2, 5, 12, -3], "MyArray").shift()),
  jsObj.console.log(jsObj.objects.array.get("MyArray").delete(2)),
  jsObj.console.log(jsObj.objects.array.get("MyArray").unshift(22)),
  jsObj.console.log(jsObj.objects.array.get("MyArray")),
  jsObj.objects.array.get("MyArray").copyWithin().setVar("youpi"),
  #jsObj.console.log(jsObj.objects.array.get("youpi").append(jsObj, 34).append(jsObj, -47)),
  #jsObj.while_()
  # jsObj.console.log(jsObj.objects.array.get("MyArray").reduce([
  #     jsObj.data.reduce.val + jsObj.data.reduce.rVal,
  #     jsObj.return_(jsObj.data.reduce.val)
  #   ]).add(2)
  #)
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
jsObj = Js.JsBase()
f = JsUtils.JsFile("Boolean", path=r"../outs")
f.writeJs([
  jsObj.objects.boolean.new(False, "testBool"),
  jsObj.console.log(jsObj.objects.boolean.get("testBool")),
  jsObj.console.log(jsObj.objects.boolean.get("testBool").valueOf()),
  jsObj.console.log(jsObj.objects.boolean.get("testBool").not_),
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
f = JsUtils.JsFile("Date", path=r"../outs")

dt = JsDate.JsDate.new("2019-05-03")
f.writeJs([dt,
  Js.JsConsole().log(dt.getDay()),
  Js.JsConsole().log(dt.getFullYear())
])


jsObj = Js.JsBase()
domObj = jsObj.createElement("a").text("Toupi 4").attr({"href": 'google', "title": 'google2'}).css({"color": "red"})
f.writeJs(domObj)
f.writeJs(jsObj.title(""))
f.writeJs(jsObj.objects.date.new("2019-01-01", varName="dateTest"))
#f.writeJs(jsObj.objects.date.get("dateTest").setMonth(7))
f.writeJs(jsObj.console.log(jsObj.objects.date.get("dateTest").addDays(jsObj, 7)))
f.writeJs(jsObj.console.log(jsObj.objects.date.get("dateTest").isWeedend))
f.writeJs(jsObj.console.log("#################"))
f.writeJs(jsObj.console.log(jsObj.objects.date.get("dateTest").getDate().add(jsObj.objects.date.get("2019-01-05"))))
f.writeJs(jsObj.console.log(jsObj.objects.date.get("dateTest").add(jsObj.objects.number.get("5"))))
f.writeJs(jsObj.console.log(jsObj.objects.date.get("dateTest").add(5)))
#f.writeJs(jsObj.console.log(jsObj.objects.date.get("dateTest").getMonth()))
f.writeJs(jsObj.console.log(jsObj.objects.date.get("dateTest").getStrTimeStamp()))
f.writeJs(jsObj.console.log(jsObj.body))
f.writeJs(jsObj.body.appendChild(domObj))

f.close(jsObj)


# ------------------------------------------------------------------------------------------------------------------
#
jsObj = Js.JsBase()
f = JsUtils.JsFile("Numbers2", path=r"../outs")
f.writeJs([
  jsObj.objects.number.new(349.673, "MyNumber"),
  jsObj.console.log(jsObj.objects.number.get("MyNumber").toPrecision(2)),
  jsObj.console.log(jsObj.objects.number.get("MyNumber").isFinite()),
  jsObj.console.log(jsObj.objects.number.get("MyNumber").toFixed()),
  jsObj.console.log(jsObj.objects.number.get("MyNumber").NEGATIVE_INFINITY),
  jsObj.console.log(jsObj.objects.number.get("MyNumber").toExponential()),
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
jsObj = Js.JsBase()
f = JsUtils.JsFile("Object", path=r"../outs")
f.writeJs([
  jsObj.objects.new({}, varName="MyObject"),
  jsObj.objects.number.new(13, varName="MyNumber"),
  jsObj.console.log(jsObj.objects.get("MyObject").isSealed()),
  jsObj.objects.get("MyObject").setattr("youpi", jsObj.objects.number.get("MyNumber")),
  jsObj.console.log(jsObj.objects.get("MyObject").isArray()),
  #jsObj.objects.get("MyObject").seal(),
  jsObj.objects.get("MyObject").setattr("test", 45),
  jsObj.console.log(jsObj.objects.get("MyObject").isSealed()),
  jsObj.console.log(jsObj.objects.get("MyObject")),
  jsObj.console.log(jsObj.objects.get("MyObject")["youpi"]),
  jsObj.objects.get("MyObject").keys().forEach([
    jsObj.console.log(jsObj.data.loop.val)]),
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
jsObj = Js.JsBase()
f = JsUtils.JsFile("String", path=r"../outs")

string = JsString.JsString.new("te st", "testStr")
f.writeJs([
#print(string.length.toStr())
  string,
  string.split(" "),
  string.parseFloat().isNaN(),
  string.endsWith("st"),
  string.charCodeAt(2),
  string.freeze(),
  JsString.JsString.new("2019-08-03", "testDate").toDate().setVar("MyDate"),
  jsObj.console.log(jsObj.objects.date.get("MyDate").addDays(jsObj, 6).getStrDate())
])
f.close(jsObj)



"""

"""

# https://bl.ocks.org/ctufts/f38ef0187f98c537d791d24fda4a6ef9
import json

from epyk.core.js import Imports
from epyk.core.html import Html

# The list of CSS classes
from epyk.core.css.styles import GrpCls


class Chart(Html.Html):
  """

  """
  name, category, callFnc = 'D3', 'Charts', 'plotChartD3'
  # _grpCls = GrpCls.CssGrpClassBase
  __reqJs = ['d3']
  references = {
    'Website': 'https://d3js.org/',
    'Website2': 'https://d3plus.org/examples/',
  }

  def __init__(self, aresObj, chartType, data, width, widthUnit, height, heightUnit, title, chartOptions, toolsbar,
               htmlCode, globalFilter, filterSensitive, dataSrc, profile):
    if chartOptions is None:
      chartOptions = {}
    self.title, self.toolsbar, self.seriesProperties, self.height = title, toolsbar, {'static': {}, 'dynamic': {}}, height
    super(Chart, self).__init__(aresObj, [], code=htmlCode, width=width, widthUnit=widthUnit, height=height, heightUnit=heightUnit, dataSrc=dataSrc)
    self.__chart = GraphFabric.CHARTS_FACTORY[self.name][chartType](aresObj, data, self.seriesProperties)
    self.__chart.data.attach(self)
    self.__chart.data._schema['out']['params'] = tuple(list(self.__chart.data._schema['out']['params']) + [self.htmlId])
    # self.__chart.data.post(('extend-dataset', self.seriesProperties, 'datasets'))
    self.__chart['chart_colors'] = json.dumps(aresObj.getColor('charts'))
    self.setSeriesColor(aresObj.getColor('charts'))
    self.css({'position': 'relative'})
    self.__chart.addAttr('position', chartOptions.get('legend', {'position': 'right'}).get('position', 'right'), ['legend'], category='options')
    self.__chart.addAttr('bottom', 20, ['layout', 'padding'], category='options')
    self.aresObj.jsGlobal.fnc('%s(%s)' % (self.__chart.jsCls, ', '.join(self.__chart.jsClsParams)), self.__chart.jsClassDefinition)
    self.colName, self.filterSensitive = list(self.__chart.data._schema['keys'])[0], filterSensitive
    if title:
      self.__chart.addAttr('text', title, ['title'], category='options')
      self.__chart.addAttr('display', True, ['title'], category='options')
    if globalFilter:
      if self._code is None:
        raise Exception("ERROR: ChartJs - %s -  Please add an htmlCode to name your filter" % chartType)

      if globalFilter is True:
        self.filter(data._jqId, self.colName)
      else:
        self.filter(**globalFilter)

  @property
  def chartId(self):
    """
    Return the Javascript variable of the chart
    """
    return "chart_%s" % self.htmlId

  @property
  def jqDiv(self):
    return "d3.select('#%s')" % self.htmlId

  @property
  def jqId(self):
    return "%s.select('svg')" % self.jqDiv

  def onDocumentLoadFnc(self): pass

  def onDocumentReady(self):
    self.ctx = []  # Just to ensure that the Structure of the chart component will not be changed in the python layer
    GraphFabric.Chart.resolveDict(dict([(key, val) for key, val in self.__chart.items() if val]), self.ctx)
    self.aresObj.jsOnLoadFnc.add('''window['%(htmlId)s_def'] = {%(chartDef)s}; %(jsChart)s
          ''' % {'htmlId': self.htmlId, 'chartDef': ", ".join(self.ctx), 'jsChart': self.jsGenerate(None)})

  def setSeriesColor(self, colors, seriesId=None, borderColors=None):
    pass

  def addAttr(self, key, val=None, tree=None, category=None, isPyData=True):
    pass

  def delAttr(self, keys, tree=None, category=None):
    pass

  def jsResize(self):
    self.addGlobalFnc('D3Resize(htmlId)', ''' 
      var container = d3.select(d3.select('#' + htmlId).node());
      var width = parseInt(container.style("width"));
      var height = parseInt(container.style("height"));
      var aspect = width / height;
    
      var svg = d3.select('#' + htmlId).select('svg').attr("viewBox", "0 0 " + width + " " + height)
        .attr("perserveAspectRatio", "xMinYMid").call(resize);  
      d3.select(window).on("resize." + container.attr("id"), resize);
    
      function resize() {
        var targetWidth = parseInt(container.style("width"));
        svg.attr("width", targetWidth);
        svg.attr("height", Math.round(targetWidth / aspect))
      }''')

  def jsRemove(self):
    return "d3.select('#%s').selectAll('svg').remove()" % self.htmlId

  def jsGenerate(self, jsData='data', jsDataKey=None, isPyData=False, jsId=None):
    if isPyData:
      jsData = json.dumps(jsData)
    if jsDataKey is not None:
      jsData = "%s.%s" % (jsData, jsDataKey)
    return '''
       window['%(htmlId)s_def'].data = %(dc)s;
       if(window['%(htmlId)s_chart'] !== undefined){window['%(htmlId)s_chart'].destroy();};
       window['%(htmlId)s_chart'] = new %(jsCls)s('%(htmlId)s', window['%(htmlId)s_def']); %(time)s;
       ''' % {'htmlId': self.htmlId, 'time': GraphFabric.Chart.jsLastUpdate(self.htmlId), 'dc': self.__chart.data.setId(jsData).getJs(filterSensitive=self.filterSensitive), 'jsCls': self.__chart.jsCls}

  def filter(self, jsId, colName, allSelected=True, filterGrp=None, operation="=", itemType="string"):
    raise NotImplemented()

  def click(self, jsFncs):
    pass

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>' % self.get_attrs(pyClassNames=self.style.get_classes())

  def setSeriesColor(self, colors, seriesId=None, borderColors=None):
    """
    :category:
    :rubric: JS
    :example:
    :dsc:

    """
    self.__chart._colors(colors, seriesId, borderColors)
    return self





class Config(Html.Html):
  name, category, callFnc = 'D3Bespoke', 'Charts', 'chartD3Bespoke'
  __reqJs = ['d3']
  __pyStyle = ['CssDivNoBorder']
  staticUrlMap = {
    'd3-lasso.min.js': 'https://bl.ocks.org/skokenes/raw/a85800be6d89c76c1ca98493ae777572/ec2ac78d8c18d3f85ef6cf7cdb77013096d42da3',
    'cloud.js': 'https://bl.ocks.org/jyucsiro/raw/767539a876836e920e38bc80d2031ba7'
  }

  def __init__(self, aresObj, script, data, width, widthUnit, height, heightUnit, title, chartOptions, toolsbar,
               htmlCode, globalFilter, filterSensitive, dataSrc, profile, url, d3Version):
    self.height = height
    super(Config, self).__init__(aresObj, [], code=htmlCode, width=width, widthUnit=widthUnit, height=height, heightUnit=heightUnit, dataSrc=dataSrc, profile=profile)
    # Override the version of D3
    if d3Version is not None:
      d3Version = '4.13.0' if d3Version == "4" else d3Version
      for mod in Imports.JS_IMPORTS['d3']['modules']:
        mod['version'] = d3Version
    # Load a external script using D3
    if script is not None:
      if url is None:
        # Try to get the script location from the catalog
        url = self.staticUrlMap.get(script)
      Imports.JS_IMPORTS[script] = {'req': [{'alias': 'd3'}], 'modules': [{'reqAlias': 'bootstrap', 'script': script, 'version': '', 'path': '', 'cdnjs': url}]}
      self.aresObj.jsImports.add(script)

  @property
  def jqId(self): return "d3.select('#%s')" % self.htmlId

  @property
  def dataId(self): return "data_%s" % self.htmlId

  def onDocumentLoadFnc(self): pass

  def dataText(self, dataText, dataType="json"):
    if dataType == 'json':
      data = json.loads(dataText)
      self.aresObj.jsOnLoadFnc.add("var %s = %s" % (self.dataId, json.dumps(data)))
    elif dataType == 'csv':
      rows, records = dataText.strip().split("\n"), []
      header = rows[0].strip().split(",")
      for row in rows[1:]:
        records.append(dict(zip(header, row.strip().split(","))))
      self.aresObj.jsOnLoadFnc.add("var %s = %s" % (self.dataId, json.dumps(records)))
    else:
      self.aresObj.jsOnLoadFnc.add("var %s = %s" % (self.dataId, json.dumps(dataText)))
    return self

  def cssText(self, cssText):
    self.aresObj.addCssText(cssText)
    return self

  def jsText(self, jsText):
    self.aresObj.jsOnLoadFnc.add(jsText)
    return self

  def jsGenerate(self, jsData='data', jsDataKey=None, isPyData=False, jsId=None):
    return ""

  def onDocumentReady(self): pass

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<svg %s></svg>' % self.get_attrs(pyClassNames=self.style.get_classes())






"""
Related Pages:

		https://www.w3schools.com/charsets/ref_utf_geometric.asp
"""

BLACK_SQUARE = '&#9632'
WHITE_SQUARE = '&#9633'
WHITE_SQUARE_WITH_ROUNDED_CORNERS = '&#9634'
WHITE_SQUARE_CONTAINING_BLACK_SMALL_SQUARE = '&#9635'
SQUARE_WITH_HORIZONTAL_FILL = '&#9636'
SQUARE_WITH_VERTICAL_FILL = '&#9637'
SQUARE_WITH_ORTHOGONAL_CROSSHATCH_FILL = '&#9638'
SQUARE_WITH_UPPER_LEFT_TO_LOWER_RIGHT_FILL = '&#9639'
SQUARE_WITH_UPPER_RIGHT_TO_LOWER_LEFT_FILL = '&#9640'
SQUARE_WITH_DIAGONAL_CROSSHATCH_FILL = '&#9641'
BLACK_SMALL_SQUARE = '&#9642'
WHITE_SMALL_SQUARE = '&#9643'
BLACK_RECTANGLE = '&#9644'
WHITE_RECTANGLE = '&#9645'
BLACK_VERTICAL_RECTANGLE = '&#9646'
WHITE_VERTICAL_RECTANGLE = '&#9647'
BLACK_PARALLELOGRAM = '&#9648'
WHITE_PARALLELOGRAM = '&#9649'
BLACK_UP_POINTING_TRIANGLE = '&#9650'
WHITE_UP_POINTING_TRIANGLE = '&#9651'
BLACK_UP_POINTING_SMALL_TRIANGLE = '&#9652'
WHITE_UP_POINTING_SMALL_TRIANGLE = '&#9653'
BLACK_RIGHT_POINTING_TRIANGLE = '&#9654'
WHITE_RIGHT_POINTING_TRIANGLE = '&#9655'
BLACK_RIGHT_POINTING_SMALL_TRIANGLE = '&#9656'
WHITE_RIGHT_POINTING_SMALL_TRIANGLE = '&#9657'
BLACK_RIGHT_POINTING_POINTER = '&#9658'
WHITE_RIGHT_POINTING_POINTER = '&#9659'
BLACK_DOWN_POINTING_TRIANGLE = '&#9660'
WHITE_DOWN_POINTING_TRIANGLE = '&#9661'
BLACK_DOWN_POINTING_SMALL_TRIANGLE = '&#9662'
WHITE_DOWN_POINTING_SMALL_TRIANGLE = '&#9663'
BLACK_LEFT_POINTING_TRIANGLE = '&#9664'
WHITE_LEFT_POINTING_TRIANGLE = '&#9665'
BLACK_LEFT_POINTING_SMALL_TRIANGLE = '&#9666'
WHITE_LEFT_POINTING_SMALL_TRIANGLE = '&#9667'
BLACK_LEFT_POINTING_POINTER = '&#9668'
WHITE_LEFT_POINTING_POINTER = '&#9669'
BLACK_DIAMOND = '&#9670'
WHITE_DIAMOND = '&#9671'
WHITE_DIAMOND_CONTAINING_BLACK_SMALL_DIAMOND = '&#9672'
FISHEYE = '&#9673'
LOZENGE = '&#9674'
WHITE_CIRCLE = '&#9675'
DOTTED_CIRCLE = '&#9676'
CIRCLE_WITH_VERTICAL_FILL = '&#9677'
BULLSEYE = '&#9678'
BLACK_CIRCLE = '&#9679'
CIRCLE_WITH_LEFT_HALF_BLACK = '&#9680'
CIRCLE_WITH_RIGHT_HALF_BLACK = '&#9681'
CIRCLE_WITH_LOWER_HALF_BLACK = '&#9682'
CIRCLE_WITH_UPPER_HALF_BLACK = '&#9683'
CIRCLE_WITH_UPPER_RIGHT_QUADRANT_BLACK = '&#9684'
CIRCLE_WITH_ALL_BUT_UPPER_LEFT_QUADRANT_BLACK = '&#9685'
LEFT_HALF_BLACK_CIRCLE = '&#9686'
RIGHT_HALF_BLACK_CIRCLE = '&#9687'
INVERSE_BULLET = '&#9688'
INVERSE_WHITE_CIRCLE = '&#9689'
UPPER_HALF_INVERSE_WHITE_CIRCLE = '&#9690'
LOWER_HALF_INVERSE_WHITE_CIRCLE = '&#9691'
UPPER_LEFT_QUADRANT_CIRCULAR_ARC = '&#9692'
UPPER_RIGHT_QUADRANT_CIRCULAR_ARC = '&#9693'
LOWER_RIGHT_QUADRANT_CIRCULAR_ARC = '&#9694'
LOWER_LEFT_QUADRANT_CIRCULAR_ARC = '&#9695'
UPPER_HALF_CIRCLE = '&#9696'
LOWER_HALF_CIRCLE = '&#9697'
BLACK_LOWER_RIGHT_TRIANGLE = '&#9698'
BLACK_LOWER_LEFT_TRIANGLE = '&#9699'
BLACK_UPPER_LEFT_TRIANGLE = '&#9700'
BLACK_UPPER_RIGHT_TRIANGLE = '&#9701'
WHITE_BULLET = '&#9702'
SQUARE_WITH_LEFT_HALF_BLACK = '&#9703'
SQUARE_WITH_RIGHT_HALF_BLACK = '&#9704'
SQUARE_WITH_UPPER_LEFT_DIAGONAL_HALF_BLACK = '&#9705'
SQUARE_WITH_LOWER_RIGHT_DIAGONAL_HALF_BLACK = '&#9706'
WHITE_SQUARE_WITH_VERTICAL_BISECTING_LINE = '&#9707'
WHITE_UP_POINTING_TRIANGLE_WITH_DOT = '&#9708'
UP_POINTING_TRIANGLE_WITH_LEFT_HALF_BLACK = '&#9709'
UP_POINTING_TRIANGLE_WITH_RIGHT_HALF_BLACK = '&#9710'
LARGE_CIRCLE = '&#9711'
WHITE_SQUARE_WITH_UPPER_LEFT_QUADRANT = '&#9712'
WHITE_SQUARE_WITH_LOWER_LEFT_QUADRANT = '&#9713'
WHITE_SQUARE_WITH_LOWER_RIGHT_QUADRANT = '&#9714'
WHITE_SQUARE_WITH_UPPER_RIGHT_QUADRANT = '&#9715'
WHITE_CIRCLE_WITH_UPPER_LEFT_QUADRANT = '&#9716'
WHITE_CIRCLE_WITH_LOWER_LEFT_QUADRANT = '&#9717'
WHITE_CIRCLE_WITH_LOWER_RIGHT_QUADRANT = '&#9718'
WHITE_CIRCLE_WITH_UPPER_RIGHT_QUADRANT = '&#9719'
UPPER_LEFT_TRIANGLE = '&#9720'
UPPER_RIGHT_TRIANGLE = '&#9721'
LOWER_LEFT_TRIANGLE = '&#9722'
WHITE_MEDIUM_SQUARE = '&#9723'
BLACK_MEDIUM_SQUARE = '&#9724'
WHITE_MEDIUM_SMALL_SQUARE = '&#9725'
BLACK_MEDIUM_SMALL_SQUARE = '&#9726'
LOWER_RIGHT_TRIANGLE = '&#9727'


"""
Module in charge of the Basic CSS events
"""

from epyk.core.css.styles.classes import CssStyle


class CssDivOnHover(CssStyle.Style):
  """
  Change the color when the mouse is on the component
  """
  _hover = {"cursor": 'pointer'}

  def customize(self):
    self.hover.css({"color": self.rptObj.theme.colors[1]}, important=True)


class CssDivOnHoverBackgroundLight(CssStyle.Style):
  """
  Change the background color when the mouse is on the component
  """
  _attrs = {"background-color": "inherit"}
  _hover = {"cursor": 'pointer'}

  def customize(self):
    self.hover.css({"background-color": self.rptObj.theme.greys[2], "font-weight": 'bold'})
    self.hover.css('color', self.rptObj.theme.greys[-1], important=True)


class CssDivOnHoverWidth(CssStyle.Style):
  """
  Change the background color when the mouse is on the component
  """
  _attrs = {"align": 'right', "background-color": "inherit"}
  _hover = {"cursor": 'pointer', 'width': '150px'}

  def customize(self):
    self.hover.css({"background-color": self.rptObj.theme.greys[2], "font-weight": 'bold'})
    self.hover.css('color', self.rptObj.theme.greys[-1], important=True)




from epyk.core.html import Html
from epyk.core.js.packages import JsQuery

# The list of CSS classes
from epyk.core.css.styles import GrpChart


class Sparklines(Html.Html):
  __reqJs = ['jquery-sparklines']
  name, category, callFnc = 'sparkline', 'Charts', 'sparkline'

  def __init__(self, report, data, chart_type, title, options):
    super(Sparklines, self).__init__(report, data)
    self._jsStyles, self.title = {"type": chart_type}, None
    if title is not None:
      self.title = self._report.ui.title(title, level=4)
      self.title.inReport = False
    if options is not None:
      self._jsStyles.update(options)

  @property
  def style(self):
    """

    :rtype: GrpChart.ClassBSpartlines
    """
    if self._styleObj is None:
      self._styleObj = GrpChart.ClassBSpartlines(self)
    return self._styleObj

  def options(self, attrs):
    """
    Add any option to the sparkline object

    Example
    chartObj.options({"lineColor": 'yellow'})

    Related Pages:

			https://omnipotent.net/jquery.sparkline/#s-docs

    :param attrs: A python dictionary

    :return: The sparkline object
    """
    self._jsStyles.update(attrs)
    return self

  @property
  def _js__builder__(self):
    return '%s.sparkline(data, options)' % JsQuery.decorate_var("htmlObj", convert_var=False)

  def __str__(self):
    # Javascript builder is mandatory for this object
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    if self.title is not None:
      return "<div style='display:inline-block;text-align:center'>%s<span %s></span></div>" % (self.title, self.get_attrs(pyClassNames=self.style.get_classes()))

    return "<span %s>Loading..</span>" % self.get_attrs(pyClassNames=self.style.get_classes())



"""
CSS Style module for the Bespoke Excel components
"""

from epyk.core.css.styles.classes import CssStyle


class CssTableExcel(CssStyle.Style):
  _attrs = {'border-collapse': 'collapse', 'border-spacing': 0, 'margin': 0, 'border': '1px solid #e5e0e0'}
  _selectors = {'child': 'table'}


class CssTableExcelHeaderCell(CssStyle.Style):
  _attrs = {'border': '1px solid #cecece', 'padding': '1px 5px', 'text-align': 'center', 'background-color': '#DCDCDC'}
  _selectors = {'child': 'table th'}


class CssTableExcelTd(CssStyle.Style):
  _attrs = {'padding': '0', 'margin': 0}
  _selectors = {'child': 'table td'}


class CssTableExcelCell(CssStyle.Style):
  _attrs = {'border': '2px solid white', 'background-color': 'white', 'height': '100%', 'width': '100%'}
  _selectors = {'child': 'td > div'}
  _active = {'border': '2px solid #8EB0E7'}


class CssTableExcelTitle(CssStyle.Style):
  _attrs = {'border': '1px solid #cecece', 'cursor': 'pointer', 'padding': '1px 20px 1px 20px',
           'background-color': '#F3F3F3'}
  _hover, _active = {'cursor': 'pointer'}, {'background-color': '#8EB0E7'}
  _selectors = {'child': 'td.rows'}


# -------------------------------------------------------------------------------------------
# CSS Style for the selected rows
# -------------------------------------------------------------------------------------------
class CssTableExcelSelectedRow(CssStyle.Style):
  _attrs = {'background-color': '#E6EFFF'}
  _selectors = {'child': 'tr.blue td > div'}


class CssTableExcelSelected(CssStyle.Style):
  _attrs = {'background-color': '#8EB0E7', 'border': '1px dotted #5292F7'}
  _selectors = {'child': 'tr.blue td:first-child'}


class CssTableRedCells(CssStyle.Style):
  _attrs = {'color': 'red'}
  _after = {'content': '" $"'}


class CssTableBackGroundRedCells(CssStyle.Style):
  _attrs = {'color': 'red'}




from epyk.core.js.html import JsHtml

from epyk.core.js.primitives import JsObjects


class Radio(JsHtml.JsHtmlRich):

  @property
  def val(self):
    """

    :return:
    """
    return JsObjects.JsObjects.get('''{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset(), name: %s, selected: %s}}
        ''' % (self.htmlId, self.content.toStr(), self.getAttribute('name'), self.selected.toStr()))

  @property
  def content(self):
    """
    Description:
    -----------

    :return:
    """
    return JsHtml.ContentFormatters(self._report, "%s.checked" % self._src.input.dom.varName)

  @property
  def selected(self):
    """

    :return:
    """
    return JsHtml.ContentFormatters(self._report, "document.body.querySelector('input[name='+%s+']:checked').getAttribute('data-content')" % self._src.input.dom.getAttribute('name'))


class Check(JsHtml.JsHtmlRich):

  @property
  def val(self):
    """

    :return:
    """
    return JsObjects.JsObjects.get('''{%s: {value: %s, timestamp: Date.now(), offset: new Date().getTimezoneOffset(), name: %s}}
        ''' % (self.htmlId, self.content.toStr(), self.getAttribute('name')))

  @property
  def content(self):
    """
    Description:
    -----------

    :return:
    """
    return JsHtml.ContentFormatters(self._report, "%s.checked" % self.varName)


class JsHtmlFields(JsHtml.JsHtmlRich):

  @property
  def val(self):
    """
    Description:
    -----------

    :return:
    """
    return self._src.input.dom.val

  @property
  def content(self):
    """
    Description:
    -----------

    :return:
    """
    return self._src.input.dom.content

  def empty(self):
    """
    Description:
    -----------

    :return:
    """
    return JsObjects.JsObjects.get('%s = ""' % self.content.toStr())


class Textarea(JsHtml.JsHtmlRich):

  @property
  def content(self):
    """
    Description:
    -----------

    :return:
    """
    return JsHtml.ContentFormatters(self._report, "%s.value" % self.varName)




from epyk.core.css import Defaults_css
from epyk.core.css.styles.attributes import Attrs


class AttrDiv(Attrs):
  def __init__(self, rptObj):
    super(AttrDiv, self).__init__(rptObj)
    self.font_size = Defaults_css.font()
    self.vertical_align = "middle"
    self.box_sizing = 'border-box'


class AttrModal(Attrs):
  def __init__(self, rptObj):
    super(AttrModal, self).__init__(rptObj)
    self.background_color = 'rgb(0,0,0,0.4)'
    self.display = 'none'




import json

from epyk.core.js import JsUtils
from epyk.core.js.objects import JsNodeDom
from epyk.core.js.packages import JsPackage
from epyk.core.js.primitives import JsObjects
from epyk.core.js.packages import JsQuery


class JsSelectItem():

  def __init__(self, selector):
    self._selector = selector

  def css(self, attrs):
    """
    Description:
    ------------

    Related Pages:
:param attrs:
    """
    return JsObjects.JsObjects.get("%s.css(%s)" % (self._selector, attrs))

  def prop(self, name, jsData):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    :param jsData:
    """
    jsData = JsUtils.jsConvertData(jsData, None)
    return JsObjects.JsObjects.get('%s.prop("%s", %s)' % (self._selector, name, jsData))


class JSelect(JsPackage):
  lib_alias = {"js": 'jqueryui', 'css': 'jqueryui'}
  lib_selector = 'jQuery("body")'
  lib_set_var = False

  def __init__(self, htmlObj, varName=None, setVar=True, isPyData=True, report=None):
    self.htmlId = varName if varName is not None else htmlObj.htmlId
    self.varName, self.varData, self.__var_def = "document.getElementById('%s')" % self.htmlId, "", None
    self._src, self._report = htmlObj, report
    self._js, self._jquery = [], None

  def val(self, jsData=None):
    """
    Description:
    ------------

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/

    Attributes:
    ----------
    :param jsData:
    """
    if jsData is None:
      return JsObjects.JsObjects.get("%s.val()" % self._src.dom.jquery.varId)

    jsData = JsUtils.jsConvertData(jsData, None)
    return JsObjects.JsObjects.get("%s.val(%s).selectpicker('refresh')" % (self._src.dom.jquery.varId, jsData))

  def prop(self, name, jsData):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param name:
    :param jsData:
    """
    jsData = JsUtils.jsConvertData(jsData, None)
    return JsObjects.JsObjects.get('%s.prop("%s", %s)' % (self._src.dom.jquery.varId, name, jsData))

  def empty(self):
    """
    Description:
    ------------

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/
    """
    if 'multiple' in self._src.attr:
      return JsObjects.JsObjects.get("%s.val([]).selectpicker('refresh')" % (self._src.dom.jquery.varId))

    return JsObjects.JsObjects.get("%s.val('').selectpicker('refresh')" % (self._src.dom.jquery.varId))

  def remove(self, jsData, refresh=True):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsData:
    :param refresh:
    """
    jsData = JsUtils.jsConvertData(jsData, None)
    if refresh:
      return JsObjects.JsObjects.get('%s.find("option[value="+ %s +"]").remove(); %s' % (self._src.dom.jquery.varId, jsData, self.refresh()))

    return JsObjects.JsObjects.get('%s.find("option[value="+ %s +"]").remove()' % (self._src.dom.jquery.varId, jsData))

  def item(self, value):
    """

    Related Pages:
:param value:

    :return:
    """
    return JsSelectItem('%s.find("option[value=Chicago]")' % self._src.dom.jquery.varId)

  def deselectAll(self):
    """
    Description:
    ------------
    This will deselect all items in a multi-select.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/
    """
    return JsObjects.JsObjects.get("%s.selectpicker('deselectAll')" % self._src.dom.jquery.varId)

  def selectAll(self):
    """
    Description:
    ------------
    This will select all items in a multi-select.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/
    """
    return JsObjects.JsObjects.get("%s.selectpicker('selectAll')" % self._src.dom.jquery.varId)

  def render(self):
    """
    Description:
    ------------
    You can force a re-render of the bootstrap-select ui with the render method.
    This is useful if you programatically change any underlying values that affect the layout of the element.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/
    """
    return JsObjects.JsObjects.get("%s.selectpicker('render')" % self._src.dom.jquery.varId)

  def refresh(self):
    """
    Description:
    ------------

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/
    """
    return JsObjects.JsObjects.get("%s.selectpicker('refresh')" % self._src.dom.jquery.varId)

  def toogle(self):
    """
    Description:
    ------------
    Programmatically toggles the bootstrap-select menu open/closed.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/
    """
    return JsObjects.JsObjects.get("%s.selectpicker('toggle')" % self._src.dom.jquery.varId)

  def hide(self):
    """
    Description:
    ------------
    To programmatically hide the bootstrap-select use the hide method (this only affects the visibility of the bootstrap-select itself).

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/
    """
    return JsObjects.JsObjects.get("%s.selectpicker('hide')" % self._src.dom.jquery.varId)

  def show(self):
    """
    Description:
    ------------
    To programmatically show the bootstrap-select use the show method (this only affects the visibility of the bootstrap-select itself).

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/
    """
    return JsObjects.JsObjects.get("%s.selectpicker('show')" % self._src.dom.jquery.varId)

  def setStyle(self, class_name, event=None):
    """
    Description:
    ------------
    To programmatically show the bootstrap-select use the show method (this only affects the visibility of the bootstrap-select itself).

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/

    Attributes:
    ----------
    :param class_name:
    :param event:
    """
    if event is None:
      return JsObjects.JsObjects.get("%s.selectpicker('setStyle', '%s')" % (self._src.dom.jquery.varId, class_name))

    return JsObjects.JsObjects.get("%s.selectpicker('setStyle', '%s', '%s')" % (self._src.dom.jquery.varId, class_name, event))

  def disable(self, flag):
    """
    Description:
    ------------

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/methods/

    Attributes:
    ----------
    :param flag:
    """
    flag = JsUtils.jsConvertData(flag, None)
    return JsObjects.JsObjects.get("%s.prop('disabled', %s)" % (self._src.dom.jquery.varId, flag))

  @property
  def search(self):
    """
    Description:
    ------------
    Get the selected content from the Select component

    Usage::

      b.click([rptObj.js.console.log(s.dom.content)])
    """
    return JsObjects.JsObjects.get("this.plugin.query")

  @property
  def events(self):
    """
    Description:
    ------------

    """
    return JsNodeDom.JsDomEvents(self._src)

  @property
  def jquery(self):
    """
    Description:
    ------------
    jQuery UI is a curated set of user interface interactions, effects, widgets, and themes built on top of the jQuery JavaScript Library.
    Whether you're building highly interactive web applications or you just need to add a date picker to a form control, jQuery UI is the perfect choice.

    Related Pages:

			https://jqueryui.com/

    :rtype: JsQuery.JQuery
    """
    if self._jquery is None:
      self._jquery = JsQuery.JQuery(src=self._src, selector=JsQuery.decorate_var("#%s" % self._src.htmlId))
    return self._jquery

  def ajaxSelectPicker(self, options):
    """
    Description:
    ------------
    Process the raw data returned from a request. The following arguments are passed to this callback:

    Related Pages:

			https://github.com/truckingsim/Ajax-Bootstrap-Select

    Usage::

      s.dom.ajaxSelectPicker({"values": "a, b, c", "ajax": {
    "url": 'https://jsonplaceholder.typicode.com/posts', "type": "POST", "dataType": "json",
    "data": {"q": "{{{q}}}"}}, 'preserveSelected': False, 'log': 2, 'preprocessData':
        'function(data) {return [{text: "C", value: "C"}, {text: "D", value: "D"}]}', "locale": {
    "emptyTitle": "Select and Begin Typing"}})])

    Attributes:
    ----------
    :param options:
    """
    self._src.cssImport.add('select-ajax')
    self._src.jsImports.add('select-ajax')
    opts = []
    for k, v in options.items():
      if not isinstance(v, (dict, int)) and v.startswith("function"):
        opts.append("%s: %s" % (k, v))
      elif isinstance(v, dict):
        tmp_dict = []
        for sk, sv in v.items():
          if not isinstance(sv, (dict, int)) and (sv.startswith("function") or sv.startswith("(function") or sv.startswith('"')):
            tmp_dict.append("%s: %s" % (sk, sv))
          else:
            tmp_dict.append("%s: %s" % (sk, json.dumps(sv)))
        opts.append("%s: {%s}" % (k, ", ".join(tmp_dict)))
      else:
        opts.append("%s: %s" % (k, json.dumps(v)))
    return "%s.selectpicker().ajaxSelectPicker({%s})" % (self.jquery.varId, ", ".join(opts))

  def __str__(self):
    """
    Description:
    ------------
    The str() method return the variable Javascript reference of the variable.

    According to the variable definition it can be either its name or its value

    :return: A Javascript reference
    """
    return self.htmlId



"""

"""

from epyk.core.js import Js
from epyk.core.js import JsUtils
from epyk.tests import test_statics

# Create a blank Javascript Python object
jsObj = Js.JsBase()

# Define a dataSet using Vis library
dataset = jsObj.data.dataset(data=[{"column": 200}], var_name="test")
opts = dataset.options()
opts.height(100)
print(opts)
dataset.setOptions(opts)

jsObj.data.dataview([])
# Write the results to a Javascript file with a Launcher
f = JsUtils.JsFile("DataSet", path=test_statics.OUTPUT_PATHS)

# Write the Javascript fragments to the file
f.writeJs([
  dataset.toStr(),
])

# Close the file and print the location of the launcher
print(f.close(jsObj))


"""
Related Pages:

		https://www.w3schools.com/html/html_symbols.asp
  https://www.w3schools.com/charsets/ref_utf_math.asp
"""

FOR_ALL = '&#8704'

COMPLEMENT = '&#8705'

PARTIAL_DIFFERENTIAL = '&#8706'

THERE_EXISTS = '&#8707'

THERE_DOES_NOT_EXIST = '&#8708'

EMPTY_SET = '&#8709'

INCREMENT = '&#8710'

NABLA = '&#8711'

ELEMENT_OF = '&#8712'

NOT_AN_ELEMENT_OF = '&#8713'

SMALL_ELEMENT_OF = '&#8714'

CONTAINS_AS_MEMBER = '&#8715'

DOES_NOT_CONTAIN_AS_MEMBER = '&#8716'

SMALL_CONTAINS_AS_MEMBER = '&#8717'

END_OF_PROOF = '&#8718'

N_ARY_PRODUCT = '&#8719'

N_ARY_COPRODUCT = '&#8720'

N_ARY_SUMMATION = '&#8721'

MINUS_SIGN = '&#8722'

MINUS_OR_PLUS_SIGN = '&#8723'

DOT_PLUS = '&#8724'

DIVISION_SLASH = '&#8725'

SET_MINUS = '&#8726'

ASTERISK_OPERATOR = '&#8727'

RING_OPERATOR = '&#8728'

BULLET_OPERATOR = '&#8729'

SQUARE_ROOT = '&#8730'

CUBE_ROOT = '&#8731'

FOURTH_ROOT = '&#8732'

PROPORTIONAL_TO = '&#8733'

INFINITY = '&#8734'

RIGHT_ANGLE = '&#8735'

ANGLE = '&#8736'

MEASURED_ANGLE = '&#8737'

SPHERICAL_ANGLE = '&#8738'

DIVIDES = '&#8739'

DOES_NOT_DIVIDE = '&#8740'

PARALLEL_TO = '&#8741'

NOT_PARALLEL_TO = '&#8742'

LOGICAL_AND = '&#8743'

LOGICAL_OR = '&#8744'

INTERSECTION = '&#8745'

UNION = '&#8746'

INTEGRAL = '&#8747'

DOUBLE_INTEGRAL = '&#8748'

TRIPLE_INTEGRAL = '&#8749'

CONTOUR_INTEGRAL = '&#8750'

SURFACE_INTEGRAL = '&#8751'

VOLUME_INTEGRAL = '&#8752'

CLOCKWISE_INTEGRAL = '&#8753'

CLOCKWISE_CONTOUR_INTEGRAL = '&#8754'

ANTICLOCKWISE_CONTOUR_INTEGRAL = '&#8755'

THEREFORE = '&#8756'

BECAUSE = '&#8757'

RATIO = '&#8758'

PROPORTION = '&#8759'

DOT_MINUS = '&#8760'

EXCESS = '&#8761'

GEOMETRIC_PROPORTION = '&#8762'

HOMOTHETIC = '&#8763'

TILDE_OPERATOR = '&#8764'

REVERSED_TILDE = '&#8765'

INVERTED_LAZY_S = '&#8766'

SINE_WAVE = '&#8767'

WREATH_PRODUCT = '&#8768'

NOT_TILDE = '&#8769'

MINUS_TILDE = '&#8770'

ASYMPTOTICALLY_EQUAL_TO = '&#8771'

NOT_ASYMPTOTICALLY_EQUAL_TO = '&#8772'

APPROXIMATELY_EQUAL_TO = '&#8773'

APPROXIMATELY_BUT_NOT_ACTUALLY_EQUAL_TO = '&#8774'

NEITHER_APPROXIMATELY_NOR_ACTUALLY_EQUAL_TO = '&#8775'

ALMOST_EQUAL_TO = '&#8776'

NOT_ALMOST_EQUAL_TO = '&#8777'

ALMOST_EQUAL_OR_EQUAL_TO = '&#8778'

TRIPLE_TILDE = '&#8779'

ALL_EQUAL_TO = '&#8780'

EQUIVALENT_TO = '&#8781'

GEOMETRICALLY_EQUIVALENT_TO = '&#8782'

DIFFERENCE_BETWEEN = '&#8783'

APPROACHES_THE_LIMIT = '&#8784'

GEOMETRICALLY_EQUAL_TO = '&#8785'

APPROXIMATELY_EQUAL_TO_OR_THE_IMAGE_OF = '&#8786'

IMAGE_OF_OR_APPROXIMATELY_EQUAL_TO = '&#8787'

COLON_EQUALS = '&#8788'

EQUALS_COLON = '&#8789'

RING_IN_EQUAL_TO = '&#8790'

RING_EQUAL_TO = '&#8791'

CORRESPONDS_TO = '&#8792'

ESTIMATES = '&#8793'

EQUIANGULAR_TO = '&#8794'

STAR_EQUALS = '&#8795'

DELTA_EQUAL_TO = '&#8796'

EQUAL_TO_BY_DEFINITION = '&#8797'

MEASURED_BY = '&#8798'

QUESTIONED_EQUAL_TO = '&#8799'

NOT_EQUAL_TO = '&#8800'

IDENTICAL_TO = '&#8801'

NOT_IDENTICAL_TO = '&#8802'

STRICTLY_EQUIVALENT_TO = '&#8803'

LESS_THAN_OR_EQUAL_TO = '&#8804'

GREATER_THAN_OR_EQUAL_TO = '&#8805'

LESS_THAN_OVER_EQUAL_TO = '&#8806'

GREATER_THAN_OVER_EQUAL_TO = '&#8807'

LESS_THAN_BUT_NOT_EQUAL_TO = '&#8808'

GREATER_THAN_BUT_NOT_EQUAL_TO = '&#8809'

MUCH_LESS_THAN = '&#8810'

MUCH_GREATER_THAN = '&#8811'

BETWEEN = '&#8812'

NOT_EQUIVALENT_TO = '&#8813'

NOT_LESS_THAN = '&#8814'

NOT_GREATER_THAN = '&#8815'

NEITHER_LESS_THAN_NOR_EQUAL_TO = '&#8816'

NEITHER_GREATER_THAN_NOR_EQUAL_TO = '&#8817'

LESS_THAN_OR_EQUIVALENT_TO = '&#8818'

GREATER_THAN_OR_EQUIVALENT_TO = '&#8819'

NEITHER_LESS_THAN_NOR_EQUIVALENT_TO = '&#8820'

NEITHER_GREATER_THAN_NOR_EQUIVALENT_TO = '&#8821'

LESS_THAN_OR_GREATER_THAN = '&#8822'

GREATER_THAN_OR_LESS_THAN = '&#8823'

NEITHER_LESS_THAN_NOR_GREATER_THAN = '&#8824'

NEITHER_GREATER_THAN_NOR_LESS_THAN = '&#8825'

PRECEDES = '&#8826'

SUCCEEDS = '&#8827'

PRECEDES_OR_EQUAL_TO = '&#8828'

SUCCEEDS_OR_EQUAL_TO = '&#8829'

PRECEDES_OR_EQUIVALENT_TO = '&#8830'

SUCCEEDS_OR_EQUIVALENT_TO = '&#8831'

DOES_NOT_PRECEDE = '&#8832'

DOES_NOT_SUCCEED = '&#8833'

SUBSET_OF = '&#8834'

SUPERSET_OF = '&#8835'

NOT_A_SUBSET_OF = '&#8836'

NOT_A_SUPERSET_OF = '&#8837'

SUBSET_OF_OR_EQUAL_TO = '&#8838'

SUPERSET_OF_OR_EQUAL_TO = '&#8839'

NEITHER_A_SUBSET_OF_NOR_EQUAL_TO = '&#8840'

NEITHER_A_SUPERSET_OF_NOR_EQUAL_TO = '&#8841'

SUBSET_OF_WITH_NOT_EQUAL_TO = '&#8842'

SUPERSET_OF_WITH_NOT_EQUAL_TO = '&#8843'

MULTISET = '&#8844'

MULTISET_MULTIPLICATION = '&#8845'

MULTISET_UNION = '&#8846'

SQUARE_IMAGE_OF = '&#8847'

SQUARE_ORIGINAL_OF = '&#8848'

SQUARE_IMAGE_OF_OR_EQUAL_TO = '&#8849'

SQUARE_ORIGINAL_OF_OR_EQUAL_TO = '&#8850'

SQUARE_CAP = '&#8851'

SQUARE_CUP = '&#8852'

CIRCLED_PLUS = '&#8853'

CIRCLED_MINUS = '&#8854'

CIRCLED_TIMES = '&#8855'

CIRCLED_DIVISION_SLASH = '&#8856'

CIRCLED_DOT_OPERATOR = '&#8857'

CIRCLED_RING_OPERATOR = '&#8858'

CIRCLED_ASTERISK_OPERATOR = '&#8859'

CIRCLED_EQUALS = '&#8860'

CIRCLED_DASH = '&#8861'

SQUARED_PLUS = '&#8862'

SQUARED_MINUS = '&#8863'

SQUARED_TIMES = '&#8864'

SQUARED_DOT_OPERATOR = '&#8865'

RIGHT_TACK = '&#8866'

LEFT_TACK = '&#8867'

DOWN_TACK = '&#8868'

UP_TACK = '&#8869'

ASSERTION = '&#8870'

MODELS = '&#8871'

TRUE = '&#8872'

FORCES = '&#8873'

TRIPLE_VERTICAL_BAR_RIGHT_TURNSTILE = '&#8874'

DOUBLE_VERTICAL_BAR_DOUBLE_RIGHT_TURNSTILE = '&#8875'

DOES_NOT_PROVE = '&#8876'

NOT_TRUE = '&#8877'

DOES_NOT_FORCE = '&#8878'

NEGATED_DOUBLE_VERTICAL_BAR_DOUBLE_RIGHT_TURNSTILE = '&#8879'

PRECEDES_UNDER_RELATION = '&#8880'

SUCCEEDS_UNDER_RELATION = '&#8881'

NORMAL_SUBGROUP_OF = '&#8882'

CONTAINS_AS_NORMAL_SUBGROUP = '&#8883'

NORMAL_SUBGROUP_OF_OR_EQUAL_TO = '&#8884'

CONTAINS_AS_NORMAL_SUBGROUP_OR_EQUAL_TO = '&#8885'

ORIGINAL_OF = '&#8886'

IMAGE_OF = '&#8887'

MULTIMAP = '&#8888'

HERMITIAN_CONJUGATE_MATRIX = '&#8889'

INTERCALATE = '&#8890'

XOR = '&#8891'

NAND = '&#8892'

NOR = '&#8893'

RIGHT_ANGLE_WITH_ARC = '&#8894'

RIGHT_TRIANGLE = '&#8895'

N_ARY_LOGICAL_AND = '&#8896'

N_ARY_LOGICAL_OR = '&#8897'

N_ARY_INTERSECTION = '&#8898'

N_ARY_UNION = '&#8899'

DIAMOND_OPERATOR = '&#8900'

DOT_OPERATOR = '&#8901'

STAR_OPERATOR = '&#8902'

DIVISION_TIMES = '&#8903'

BOWTIE = '&#8904'

LEFT_NORMAL_FACTOR_SEMIDIRECT_PRODUCT = '&#8905'

RIGHT_NORMAL_FACTOR_SEMIDIRECT_PRODUCT = '&#8906'

LEFT_SEMIDIRECT_PRODUCT = '&#8907'

RIGHT_SEMIDIRECT_PRODUCT = '&#8908'

REVERSED_TILDE_EQUALS = '&#8909'

CURLY_LOGICAL_OR = '&#8910'

CURLY_LOGICAL_AND = '&#8911'

DOUBLE_SUBSET = '&#8912'

DOUBLE_SUPERSET = '&#8913'

DOUBLE_INTERSECTION = '&#8914'

DOUBLE_UNION = '&#8915'

PITCHFORK = '&#8916'

EQUAL_AND_PARALLEL_TO = '&#8917'

LESS_THAN_WITH_DOT = '&#8918'

GREATER_THAN_WITH_DOT = '&#8919'

VERY_MUCH_LESS_THAN = '&#8920'

VERY_MUCH_GREATER_THAN = '&#8921'

LESS_THAN_EQUAL_TO_OR_GREATER_THAN = '&#8922'

GREATER_THAN_EQUAL_TO_OR_LESS_THAN = '&#8923'

EQUAL_TO_OR_LESS_THAN = '&#8924'

EQUAL_TO_OR_GREATER_THAN = '&#8925'

EQUAL_TO_OR_PRECEDES = '&#8926'

EQUAL_TO_OR_SUCCEEDS = '&#8927'

DOES_NOT_PRECEDE_OR_EQUAL = '&#8928'

DOES_NOT_SUCCEED_OR_EQUAL = '&#8929'

NOT_SQUARE_IMAGE_OF_OR_EQUAL_TO = '&#8930'

NOT_SQUARE_ORIGINAL_OF_OR_EQUAL_TO = '&#8931'

SQUARE_IMAGE_OF_OR_NOT_EQUAL_TO = '&#8932'

SQUARE_ORIGINAL_OF_OR_NOT_EQUAL_TO = '&#8933'

LESS_THAN_BUT_NOT_EQUIVALENT_TO = '&#8934'

GREATER_THAN_BUT_NOT_EQUIVALENT_TO = '&#8935'

PRECEDES_BUT_NOT_EQUIVALENT_TO = '&#8936'

SUCCEEDS_BUT_NOT_EQUIVALENT_TO = '&#8937'

NOT_NORMAL_SUBGROUP_OF = '&#8938'

DOES_NOT_CONTAIN_AS_NORMAL_SUBGROUP = '&#8939'

NOT_NORMAL_SUBGROUP_OF_OR_EQUAL_TO = '&#8940'

DOES_NOT_CONTAIN_AS_NORMAL_SUBGROUP_OR_EQUAL = '&#8941'

VERTICAL_ELLIPSIS = '&#8942'

MIDLINE_HORIZONTAL_ELLIPSIS = '&#8943'

UP_RIGHT_DIAGONAL_ELLIPSIS = '&#8944'

DOWN_RIGHT_DIAGONAL_ELLIPSIS = '&#8945'

ELEMENT_OF_WITH_LONG_HORIZONTAL_STROKE = '&#8946'

ELEMENT_OF_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = '&#8947'

SMALL_ELEMENT_OF_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = '&#8948'

ELEMENT_OF_WITH_DOT_ABOVE = '&#8949'

ELEMENT_OF_WITH_OVERBAR = '&#8950'

SMALL_ELEMENT_OF_WITH_OVERBAR = '&#8951'

ELEMENT_OF_WITH_UNDERBAR = '&#8952'

ELEMENT_OF_WITH_TWO_HORIZONTAL_STROKES = '&#8953'

CONTAINS_WITH_LONG_HORIZONTAL_STROKE = '&#8954'

CONTAINS_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = '&#8955'

SMALL_CONTAINS_WITH_VERTICAL_BAR_AT_END_OF_HORIZONTAL_STROKE = '&#8956'

CONTAINS_WITH_OVERBAR = '&#8957'

SMALL_CONTAINS_WITH_OVERBAR = '&#8958'

Z_NOTATION_BAG_MEMBERSHIP = '&#8959'




from epyk.core.js.primitives import JsString
from epyk.core.js.primitives import JsObjects
from epyk.core.js import JsUtils


def packageImport(jsPackage=None, cssPackage=None):
  """
  Description
  ---------------
  Simple decorator to allow people to declare packages that need to be imported when they are manipulating HTML compponents
  The alias for the package needs to be defined in the Import.py module

  Usage::

      import epyk.core.js.packages
  @packageImport('myJsPackage', 'myCssPackage')
  def myHtmlComponent()
    pass
  """
  def wrap(func):
    def inner(rptObj, *args, **kwargs):
      report = getattr(rptObj, '_report') if getattr(rptObj, '_report', None) else rptObj
      if jsPackage:
        report.jsImports.add(jsPackage)
      if cssPackage:
        report.cssImport.add(cssPackage)
      return func(rptObj, *args, **kwargs)

    return inner

  return wrap


class JsPackage(object):
  lib_alias, lib_selector, lib_set_var = None, None, True

  class __internal(object):
    # By default it will attach eveything to the body
    jqId, jsImports, cssImport = '', set([]), set([])

  def __init__(self, src=None, varName=None, selector=None, data=None, setVar=None, parent=None):
    self.src = src if src is not None else self.__internal()
    self._selector = selector if selector is not None else self.lib_selector
    self.varName, self.setVar, self._parent = varName, self.lib_set_var if setVar is None else setVar, parent
    self._data = data
    self._js, self._u, self._js_enums = [[]], {}, {} # a list of list of object definition
    self.__js_uniq_fncs = {}
    if self.lib_alias is not None:
      if 'css' in self.lib_alias:
        self.src.cssImport.add(self.lib_alias['css'])
      if 'js' in self.lib_alias:
        self.src.jsImports.add(self.lib_alias['js'])

  @property
  def varId(self):
    """
    Description:
    ------------
    The Javascript and Python reference ID
    """
    return self._selector if self.varName is None else self.varName

  def version(self, ver):
    """
    Description:
    ------------
    Change the package version number

    Usage::

      bar.chart.version("1.11.0")

    Attributes:
    ----------
    :param ver: String. The package versions example 1.11.0
    """
    self.src._props.setdefault("packages", {})[self.lib_alias] = ver
    return self

  def fnc(self, data, unique=False):
    """
    Description:
    ------------
    Base function to allow the object chain.
    THis will add the elements to the current section in the object structure
    All the items at the same level wil be chained

    Attributes:
    ----------
    :param data: String. THe Javascript fragment to be added
    :param unique: Boolean. Ensure the function is available one time in the chain. If not the last call we will present

    :return: "Self" to allow the chains
    """
    if unique:
      fnc_name = data.split("(")[0]
      if fnc_name in self.__js_uniq_fncs:
        i, j = self.__js_uniq_fncs[fnc_name]
        self._js[i].pop(j)
      self.__js_uniq_fncs[fnc_name] = (len(self._js)-1, self._js[-1])
    self._js[-1].append(data)
    return self

  def fnc_enum(self, name, data_class):
    """
    Description:
    ------------
    Base function to allow the creation of function with parameters which are list of dataclasses.
    Basically this will be then transpiled to a list of dictionary

    :param name: String. The function Name
    :param data_class: Class. The Python Data class
    """
    index = len(self._js) - 1
    if not index in self._js_enums:
      self._js_enums[index] = {name: []}
      self._js[-1].append(name)
    elif not name in self._js_enums[index]:
      self._js_enums[index][name] = []
      self._js[-1].append(name)
    self._js_enums[index][name].append(data_class(self.src))
    return self._js_enums[index][name][-1]

  def fnc_closure(self, data, checkUndefined=False, unique=False):
    """
    Description:
    ------------
    Add the function string to the existing object definition but create a new entry point for the next ones.
    This structure will allow the chain on the Javascript side but also on the Python side.

    Thanks to this Python can always keep the same structure and produce the correct Javascript definition
    There will be no chain in the Javascript side

    Attributes:
    ----------
    :param data: String. The Javascript fragment to be added
    :param checkUndefined: Boolean. Add a check on the variable definition
    :param unique: Boolean. Ensure the function is available one time in the chain. If not the last call we will present

    :return: The "self" to allow the chains
    """
    if unique:
      fnc_name = data.split("(")[0]
      if fnc_name in self.__js_uniq_fncs:
        i, j = self.__js_uniq_fncs[fnc_name]
        self._js[i].pop(j)
      self.__js_uniq_fncs[fnc_name] = (len(self._js)-1, self._js[-1])

    self._js[-1].append(data)
    if checkUndefined:
      self._u[len(self._js) - 1] = checkUndefined
    self._js.append([])
    return self

  def fnc_closure_in_promise(self, data, checkUndefined=False):
    """
    Description:
    ------------
    Base function to allow the creation of a promise.

    A Js promise is an event attached toa function which will be only executed after the function.
    In case of success the then will be triggered otherwise the exception will be caught.

    Attributes:
    ----------
    :param data: String. The Javascript fragment to be added
    :param checkUndefined: Boolean. Add a check on the variable definition

    :return: The promise
    """
    self._js[-1].append(JsObjects.JsPromise(data))
    if checkUndefined:
      self._u[len(self._js) - 1] = checkUndefined
    self._js.append([])
    return data

  @property
  def var(self):
    """
    Description:
    ------------
    Property to return the variable name as a valid pyJs object
    """
    return JsString.JsString(self.varId, isPyData=False)

  def set_var(self, flag):
    """
    Description:
    ------------
    Change the flag to define if the variable should be defined on the Javascript side.
    Default this is set to True

    Attributes:
    ----------
    :param flag: A python boolean

    :return: This in order to allow js chains
    """
    self.setVar = flag
    return self

  def getStr(self, emptyStack=True):
    """
    Description:
    ------------
    Get the current string representation for the object and remove the stack
    """
    js_stack = None
    if not emptyStack:
      js_stack = list(self._js)
    content = self.toStr()
    if not emptyStack:
      self._js = js_stack
    if not content:
      return self.varId

    return content

  def _mapVarId(self, strFnc, varId):
    """
    Description:
    ------------
    Special function used for some external packages used to fix the problem of function override.
    Indeed in Datatable row.add is used as a class method compare to the other functions used at object level

    Attributes:
    ----------
    :param strFnc: The function string
    :param varId: The object reference

    :return: The converted object reference
    """
    return varId

  def toStr(self):
    """
    Description:
    ------------
    Javascript representation

    :return: Return the Javascript String
    """
    if self._selector is None:
      raise Exception("Selector not defined, use this() or new() first")

    obj_content = []
    # TODO find better way to check emtpy js structure
    if self._js != [[]]:
      for i, js in enumerate(self._js):
        if len(js) == 0:
          continue

        fnc_frg = []
        for d in js:
          if hasattr(d, "toStr"):
            fnc_frg.append(d.toStr())
          else:
            if i in self._js_enums and d in self._js_enums[i]:
              fnc_frg.append("%s([%s])" % (d, ",".join([str(s) for s in self._js_enums[i][d]])))
            else:
              fnc_frg.append(d)
        str_fnc = ".".join(fnc_frg)
        if self.setVar:
          if str_fnc:
            str_fnc = "var %s = %s.%s" % (self.varId, self._selector, str_fnc)
          else:
            str_fnc = "var %s = %s" % (self.varId, self._selector)
          self.setVar = False
        else:
          if str_fnc:
            if i in self._u:
              # to avoid raising an error when the variable is not defined
              str_fnc = "if(%s !== undefined){%s.%s}" % (self.varId, self.varId, str_fnc)
            else:
              varId = self._mapVarId(str_fnc, self.varId)
              str_fnc = "%s.%s" % (varId, str_fnc)
          else:
            str_fnc = self.varId
        obj_content.append(str_fnc)
    else:
      if self.setVar:
        obj_content = ["var %s = %s" % (self.varId, self._selector)]
        self.setVar = False
    self._js = [[]] # empty the stack
    return "; ".join(obj_content)


class DataAttrs(object):
  def __init__(self, report, attrs=None, oprions=None):
    self._report, self.oprions, self._attrs = report, oprions, attrs or {}

  def custom(self, name, value):
    """
    Description:
    ------------
    Custom function to add a bespoke attribute to a class.

    This entry point will not be able to display any documentation but it is a shortcut to test new features.
    If the value is a Javascript object, the PyJs object must be used

    Attributes:
    ----------
    :param name: String. The key to be added to the attributes
    :param value: String or JString. The value of the defined attributes

    :return: The DataAttrs to allow the chains
    """
    self._attrs[name] = JsUtils.jsConvertData(value, None)
    return self

  def attr(self, name, value):
    """
    Description:
    ------------
    Add an attribute to the Javascript underlying dictionary

    Attributes:
    ----------
    :param name: String. The attribute name
    :param value: Object. The attribute value

    :return: "Self" to allow the chains on the Python side
    """
    self._attrs[name] = value
    return self

  def attrs(self, values):
    """
    Description:
    ------------
    Set multiple attributes to the underlying data directly from a dictionary

    Attributes:
    ----------
    :param values: Dictionary. The data to set

    :return: "Self" to allow the chains on the Python side
    """
    self._attrs.update(values)
    return self

  def __str__(self):
    """
    Description:
    ------------
    Produce the resulting string to be added to the Javascript section of the web page
    """
    return "{%s}" % ", ".join(["%s: %s" % (k, v) for k, v in self._attrs.items()])

  def toStr(self):
    """

    :return:
    """
    return "{%s}" % ", ".join(["%s: %s" % (k, JsUtils.jsConvertData(v, None)) for k, v in self._attrs.items()])




from epyk.core.js.html import JsHtml
from epyk.core.js.primitives import JsObjects


class Stars(JsHtml.JsHtmlRich):

  @property
  def val(self):
    """
    Description:
    ------------
    The Javascript data object. A dictionary with all the specific metadata attached to the component
    """
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.dataset.level, timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (self.htmlId, self.varName))

  @property
  def content(self):
    """
    Description:
    ------------
    The Javascript value of the component. This returned only a value corresponding to the state of the component
    """
    return JsHtml.ContentFormatters(self._report, "%s.dataset.level" % self.varName)




# Check if pandas is available in the current environment
# if it is the case this module can handle DataFrame directly
try:
  import pandas as pd
  has_pandas = True

except:
  has_pandas = False


from epyk.core.html import tables as html_tables
from epyk.core import js

from epyk.interfaces.tables import CompTabulator
from epyk.interfaces.tables import CompDatatable
from epyk.interfaces.tables import CompTablesPlotly
from epyk.interfaces.tables import CompTableD3
from epyk.interfaces.tables import CompAgGrid


class Tables(object):
  def __init__(self, context):
    self.context = context

  @property
  def aggrid(self):
    """
    Description:
    -----------
    ag-Grid is the industry standard for JavaScript Enterprise Applications.
    Developers using ag-Grid are building applications that would not be possible if ag-Grid did not exist.

    Related Pages:

      https://www.ag-grid.com/javascript-grid/
    """
    return CompAgGrid.AgGrid(self)

  @property
  def tabulators(self):
    """
    Description:
    -----------
    Interface to the different Tabulator configurations

    Related Pages:
http://tabulator.info/
    """
    return CompTabulator.Tabulators(self)

  @property
  def d3(self):
    """
    Description:
    -----------
    Interface to the different Tabulator configurations

    Related Pages:

			https://github.com/d3/d3/wiki/Gallery

    :rtype: CompTableD3.D3
    """
    return CompTableD3.D3(self)

  @property
  def plotlys(self):
    """
    Interface to the different Tabulator configurations

    Related Pages:
http://tabulator.info/

    :rtype: CompTabulator.Tabulators
    """
    return CompTablesPlotly.Plotly(self)

  @property
  def datatables(self):
    """
    Description:
    -----------
    Interface to the different Datatable configurations

    Related Pages:

			https://datatables.net/

    :rtype: CompDatatable.Datatables
    """
    return CompDatatable.Datatables(self)

  def pivot(self, recordSet=None, rows=None, cols=None, vals=None, title='', aggOptions=None, rendererName="Table",
            tableOptions=None, width=(100, '%'), height=(None, 'px'), htmlCode=None, dataSrc=None, helper=None, profile=False):
    """
    Description:
    -----------
    Create a HTML Pivot table

    Usage::

      Related Pages:

			https://pivottable.js.org/examples/
    https://react-pivottable.js.org/
    https://jsfiddle.net/nicolaskruchten/w86bgq9o/

    :return: The python HTML Table

    :rtype: tables.HtmlDataPivot.PivotTable
    """
    if tableOptions is None:
      tableOptions = {}
    if aggOptions is None:
      aggOptions = {'name': "Sum Agg", 'digits': 2}
    elif 'name' not in aggOptions:
      aggOptions['name'] = "Sum Agg"
    return self.context.register(html_tables.HtmlTablePivot.PivotTable(self.context.rptObj, js.AresJs.Js(self.context.rptObj,
            recordSet, profile=profile), rows, cols, vals, title, tableOptions, width, height, aggOptions, rendererName,
                                                     htmlCode, dataSrc, helper, profile))

  def datatable(self, recordSet=None, header=None, dataFncs=None, aggFnc='sum', cols=None, rows=None, title='',
            width=(100, '%'), height=(None, 'px'), options=None, htmlCode=None,
            profile=None):
    """
    Description:
    -----------
    The python interface to the javascript Datatable framework. Not all the functions have been wrapped here but you should be able to
    do the most frequent events and interactions with this component from the available function.
    Please keep in mind that the javascript is only trigger on the web browser (namely not with the Python code)

    Usage::

      Related Pages:

			https://datatables.net/reference/index
    https://datatables.net/reference/option/
    https://datatables.net/reference/option/ajax.data
    https://datatables.net/reference/option/drawCallback
    https://datatables.net/extensions/buttons/examples/initialisation/custom.html

    :rtype: html_tables.HtlmTableDatatable.DataTable
    """
    rows = rows or []
    cols = cols or []
    # if recordSet is not None and cols is None and rows is None:
    #   if isinstance(recordSet, list):
    #     recordSet = self.context.rptObj.df(recordSet)
    #   if header is not None:
    #     rows, cols = [], []
    #     for i, colType in enumerate(recordSet.dtypes):
    #       colName = recordSet.dtypes.index[i]
    #       if not colName in header and (isinstance(header, dict) and not colName in header.get('_order', [])):
    #         continue
    #
    #       if colType in ['float64', 'bool', 'int64']:
    #         cols.append(colName)
    #       else:
    #         rows.append(colName)
    #   else:
    #     rows, cols = ['_index'], []
    #     header = {'_index': {'visible': False}}
    #     for i, colType in enumerate(recordSet.dtypes):
    #       colName = recordSet.dtypes.index[i]
    #       if colType in ['float64', 'bool', 'int64']:
    #         cols.append(colName)
    #       else:
    #         rows.append(colName)
    #
    # if '_index' in rows:
    #   recordSet['_index'] = recordSet.index
    _tableOptions = {'autoWidth': True} # 'scrollX': True
    if options is None:
      options = {}
    _tableOptions.update(options)
    jsFncs, systemCols = [], {}
    if _tableOptions.get('system', {}).get('age'):
      systemCols['age'] = []
      for colName in cols:
        systemCols['age'].append("%s.age" % colName)

    if _tableOptions.get('system', {}).get('quality'):
      systemCols['quality'] = []
      for colName in cols:
        systemCols['quality'].append("%s.quality" % colName)

    if _tableOptions.get('system', {}).get('intensity'):
      systemCols['intensity'] = []
      for colName in cols:
        systemCols['intensity'].append("%s.intensity" % colName)

    #if tableTypes == 'hierarchy':
    #  rows = ['_id', '_level'] + rows

    if isinstance(aggFnc, str):
      if cols is None:
        cols = []
      jsFncs.append((aggFnc, rows, cols + systemCols.get('age', []) + systemCols.get('intensity', []) + systemCols.get('quality', [])))
    else:
      jsFncs.append(aggFnc)
    # if isinstance(recordSet, list):
    #   recordSet = self.context.rptObj.df(recordSet)
    if dataFncs is not None:
      jsFncs += dataFncs
    if header is None:
      header = {'_order': rows + cols}

    elif isinstance(header, list):
      tmpHeader = {'_order': []}
      for h in header:
        if isinstance(h, dict):
          tmpHeader[h['data']] = h.get('title', h['data'])
          tmpHeader['_order'].append(h['data'])
        else:
          tmpHeader[h] = h
          tmpHeader['_order'].append(h)
      header = tmpHeader
    if profile:
      self.context.rptObj.jsOnLoadFnc.add(self.context.rptObj.jsConsole("", isPyData=True))
      self.context.rptObj.jsOnLoadFnc.add(self.context.rptObj.jsConsole("************************", isPyData=True))
      self.context.rptObj.jsOnLoadFnc.add(self.context.rptObj.jsConsole("Profiling mode for table", isPyData=True))
    return self.context.register(html_tables.HtlmTableDatatable.DataTable(self.context.rptObj, js.AresJs.Js(self.context.rptObj, recordSet, profile=profile).fncs(jsFncs, systemCols), header,
                                                                               title, width, height, _tableOptions, htmlCode, profile))

  def tabulator(self, recordSet=None, header=None, cols=None, rows=None, title=None, align="left", width=(100, '%'),
                height=(None, 'px'), options=None, htmlCode=None, profile=None):
    """
    Description:
    -----------

    Usage::

      Related Pages:
http://tabulator.info/

    Attributes:
    ----------
    :param recordSet:
    :param header:
    :param cols:
    :param rows:
    :param title:
    :param align:
    :param width:
    :param height:
    :param options:
    :param htmlCode:
    :param profile:

    :rtype: html.HtmlDataTable.DataTabulator

    :return:
    """
    table_options_dflts = {'selectable': False, 'index': '_row', 'layout': 'fitColumns', 'pagination': 'local',
                           'paginationSize': 25, 'resizableRows': False, 'movableColumns': True}
    if height[0] is not None and height[1] == 'px':
      table_options_dflts['height'] = height[0]
      height = (height[0] + 26, height[1])
    tmp_header = {}
    if header is not None:
      # Apply a formatting to the specific columns
      if '_formats' in header:
        for col in header['_formats']:
          if not 'cols' in col:
            raise Exception("List of columns should be defined in the key cols")

          cols_format = list(col['cols'])
          del col['cols']

          for cf in cols_format:
            if cf not in header:
              header[cf] = {'field': cf, 'title': cf, 'align': 'center'}
              header[cf].update(col)
        del header['_formats']

      for k, v in header.items():
        r = {'field': k, 'title': k, 'align': 'center'}
        r.update(v)
        tmp_header[k] = r
    elif recordSet is not None and len(recordSet) > 0:
      cols = recordSet[0].keys()

    order_cols = []
    if cols is not None:
      for col in cols:
        if col not in tmp_header:
          tmp_header[col] = {'title': col, 'field': col, 'align': 'center'}
        order_cols.append(col)
    if rows is not None:
      for row in rows:
        if row not in tmp_header:
          tmp_header[row] = {'title': row, 'field': row, 'formatter': 'number', 'align': 'center'}
          if options is not None and 'digits' in options:
            tmp_header[row]['pyOptions'] = {"digits": options['digits']}
        order_cols.append(row)
    if len(order_cols) == 0:
      order_cols = tmp_header.keys()
    if options is not None:
      table_options_dflts.update(options)
    if not table_options_dflts.get('paginationSize'):
      del table_options_dflts['pagination']
      del table_options_dflts['paginationSize']

    elif 'pageLength' in table_options_dflts:
      table_options_dflts['pagination'] = 'local'
      table_options_dflts['paginationSize'] = table_options_dflts['pageLength']
      del table_options_dflts['pageLength'] # datatable parameter
    if 'cellsType' in table_options_dflts:
      for c in tmp_header:
        # Do not override existing configuration
        for k, v in table_options_dflts['cellsType'].items():
          if k not in tmp_header[c]:
            tmp_header[c][k] = v
      del table_options_dflts['cellsType']

    if 'header' in table_options_dflts:
      for k, v in table_options_dflts["header"].items():
        for h, d in tmp_header.items():
          d[k] = v
    return self.context.register(html_tables.HtmlTableTabulator.DataTabulator(self.context.rptObj, js.AresJs.Js(self.context.rptObj, recordSet, profile=profile),
              [tmp_header[c] for c in order_cols if c in tmp_header], align, title, width, height,
               table_options_dflts, htmlCode, profile))

  def config(self, htmlCode, visible=False, profile=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param htmlCode:
    :param visible:
    :param profile:
    """
    return self.context.register(
      html_tables.HtmlTableConfig.ConfigTable(self.context.rptObj, htmlCode, visible, profile))

  def basic(self, records, cols, rows, width=(100, '%'), height=(None, 'px'), htmlCode=None, options=None, profile=None):
    """
    Description:
    -----------

    Usage::

      simple_table = rptObj.ui.tables.basic(df.to_dict("records"), cols=["COL1"], rows=["COL2"])
    simple_table.add({"COL1": "Value"})

    Attributes:
    ----------
    :param records:
    :param cols:
    :param rows:
    :param width:
    :param height:
    :param htmlCode:
    :param options:
    :param profile:

    :return:
    """
    table = html_tables.HtmlTable.Bespoke(self.context.rptObj, records, cols, rows, width, height, htmlCode, options, profile)
    self.context.register(table)
    return table

  def grid(self, records, cols, rows, width=(None, '%'), height=(None, 'px'), htmlCode=None, options=None, profile=None):
    """

    Usage::

      """
    width_cells, width_rows_header = 50, 100
    for rec in records:
      for c in cols:
        if c not in rec:
          rec[c] = 0
    table = html_tables.HtmlTable.Bespoke(self.context.rptObj, records, cols, rows, width, height, htmlCode, options, profile)
    table.css({"width": "%spx" % (width_rows_header + len(cols) * width_cells)})
    # table[0][0]._vals = ""
    for i in table[1:]:
      for j in range(len(rows)):
        i[j].attr["name"] = "row_header"
        i[j].css({"width": "%spx" % width_rows_header})
      for cell in i[j:]:
        cell.attr["contenteditable"] = 'true'
        cell.css({"width": '%spx' % width_cells})
    table.style.clear()
    # Set the style for the table grid
    # table.style.add_classes.table.grid_row_header()
    table.style.add_classes.table.grid_vals()
    # table.style.add_classes.table.grid_no_header()
    self.context.register(table)
    return table




import re
import json

from epyk.core.html import Html
from epyk.core.html.options import OptButton

from epyk.core.js.html import JsHtml
from epyk.core.js import JsUtils
from epyk.core.js.statements import JsIf
from epyk.core.js.objects import JsComponents

# The list of CSS classes
from epyk.core.css.styles import GrpClsButton
from epyk.core.css import Defaults_css


class Button(Html.Html):
  name, category, callFnc = 'Button', 'buttons', 'button'

  def __init__(self, report, text, icon, width, height, htmlCode, tooltip, profile, options):
    if text is not None and not isinstance(text, list):
      text = [text]
    for obj in text:
      if hasattr(obj, 'inReport'):
        obj.inReport = False
    super(Button, self).__init__(report, text or [], code=htmlCode, profile=profile, css_attrs={"width": width, "height": height})
    self.add_icon(icon)
    if icon is not None and not text:
      self.icon.style.css.margin_right = None
    self.__options = OptButton.OptionsButton(self, options or {})
    if tooltip is not None:
      self.tooltip(tooltip)
    self.set_attrs(name="data-count", value=0)

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    self.val.append(htmlObj)
    return self

  def __getitem__(self, i):
    return self.val[i]

  @property
  def options(self):
    """
    Description:
    -----------
    Property to set all the possible object for a button

    :rtype: OptButton.OptionsButton
    """
    return self.__options

  @property
  def dom(self):
    """
    Description:
    -----------
    HTML Dom object

    :rtype: JsHtml.JsHtmlButton
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtmlButton(self, report=self._report)
    return self._dom

  @property
  def _js__builder__(self):
    return "htmlObj.innerHTML = data"

  @property
  def style(self):
    """
    Description:
    -----------
    Property to the CSS Style of the component

    Usage::

      self.style.css.margin = "5px"

    :rtype: GrpClsButton.ClassButton
    """
    if self._styleObj is None:
      self._styleObj = GrpClsButton.ClassButton(self)
    return self._styleObj

  def disable(self, background_color=None, color=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param background_color: String.
    :param color: String.

    :return: The htmlObj to allow the chaining
    """
    css_pmts = {"cursor": "not-allowed"}
    if background_color is not None:
      css_pmts["background"] = background_color
    if color is not None:
      css_pmts["color"] = color
    self.css(css_pmts)
    self.attr['disabled'] = True
    return self

  def press(self, jsPressFncs=None, jsReleaseFncs=None, profile=False):
    """
    Description:
    -----------
    Special click event to keep in memory the state of the component

    Usage::

      Attributes:
    ----------
    :param jsPressFncs:
    :param jsReleaseFncs:
    :param profile:
    """
    str_fnc = ""
    if jsPressFncs is not None:
      if not isinstance(jsPressFncs, list):
        jsPressFncs = [jsPressFncs]
      if self.options.group is not None and not self.options.multiple:
        jsPressFncs.append(self.dom.release(by_name=True))
      jsPressFncs.append(self.dom.lock(not_allowed=jsReleaseFncs is None))
      str_fnc = "if(%s == 'pointer'){%s}" % (self.dom.css('cursor'), JsUtils.jsConvertFncs(jsPressFncs, toStr=True))
    if jsReleaseFncs is not None:
      if jsPressFncs is None:
        raise Exception("Press Event must be defined")

      if not isinstance(jsReleaseFncs, list):
        jsReleaseFncs = [jsReleaseFncs]
      jsReleaseFncs.append(self.dom.release())
      str_fnc = "%s else{%s}" % (str_fnc, JsUtils.jsConvertFncs(jsReleaseFncs, toStr=True))
    return self.on("click", str_fnc, profile)

  def color(self, color):
    """
    Description:
    -----------
    Change the color of the button background when the mouse is hover

    Usage::

      rptObj.ui.buttons.remove("remove").color("blue")

    Attributes:
    ----------
    :param color: String. the color of the component (text and borders)
    """
    self.style.css.border = "1px solid %s" % color
    self.set_attrs(name="onmouseover", value="this.style.backgroundColor='%s';this.style.color='white'" % color)
    self.set_attrs(name="onmouseout", value="this.style.backgroundColor=\'white\';this.style.color=\'%s\';" % color)
    return self

  def __str__(self):
    str_div = "".join([v.html() if hasattr(v, 'html') else str(v) for v in self.val])
    return '<button %s>%s</button>' % (self.get_attrs(pyClassNames=self.style.get_classes()), str_div)

  @staticmethod
  def matchMarkDown(val):
    return val.startswith("[button ")

  @classmethod
  def jsMarkDown(self, vals):
    return '[button url=""](%s)' % vals

  @classmethod
  def to_markdown(cls, val, regExpResult, rptObj):
    res = re.search('\[button url="(.*)"\]\((.*)\)', val)
    if res is not None:
      if rptObj is not None:
        getattr(rptObj, 'button')(res.group(2)).goTo(res.group(1))
      return ["rptObj.ui.%s.button('%s').goTo('%s')" % (cls.category, res.group(2), res.group(1))]

  # No export to this component as no interactivity is expected
  def to_word(self, document): pass
  def to_xls(self, workbook, worksheet, cursor): pass
  def to_ppt(self, workbook, worksheet, cursor): pass


class Checkbox(Html.Html):
  name, category, callFnc = 'Check Box', 'Buttons', 'checkbox'
  __reqCss, __reqJs = ['font-awesome', 'bootstrap'], ['font-awesome', 'bootstrap', 'jquery']

  def __init__(self, rptObj, records, title, color, width, height, align, htmlCode, filters, tooltip, icon, options, profile):
    if rptObj.http.get(htmlCode) is not None:
      selectedVals = set(rptObj.http[htmlCode].split(","))
      for rec in records:
        if rec["value"] in selectedVals:
          rec["checked"] = True
    super(Checkbox, self).__init__(rptObj, records, htmlCode=htmlCode, css_attrs={"width": width, "height": height},
                                   globalFilter=filters, profile=profile)
    # Add the component title
    self.add_title(title, options={'content_table': False})
    self.css({'text-align': align, 'color': 'inherit' if color is None else color, 'padding': '5px'})
    self._jsStyles = {"tooltip": tooltip, "icon": icon} # fas fa-circle fas fa-check
    self.selectAll = options.get("all_selected", False)

  def tooltip(self, value, location='top'):
    self._jsStyles['tooltip'] = value
    return self

  def jsDisable(self, jsData='data', jsDataKey=None, isPyData=False, jsParse=False, jsFnc=None, reset=True):
    jsData = self._jsData(jsData, jsDataKey, jsParse, isPyData, jsFnc)
    return '''
        const ldata = %(jsData)s;
        %(jqId)s.find('span').each(function(){
          var pItem = $(this).next(); 
          if(%(reset)s == true) {pItem.removeClass("%(disableCls)s")};
          if(($(this).find("i").attr("class") !== undefined) && (%(reset)s == true) && ($(this).find("i").attr("class").includes('fas fa-times'))){$(this).html('<div style="width:16px;display:inline-block">&nbsp;</div>') };
          if (ldata.indexOf(pItem.text()) > -1){
            var bcIndex = %(breadCrumVar)s['params']['%(htmlCode)s'].indexOf(pItem.text());
            if(bcIndex > -1) {delete %(breadCrumVar)s['params']['%(htmlCode)s'].splice(bcIndex, 1)};
            $(this).html('<i class="fas fa-times %(disableCls)s"></i>'); pItem.addClass("%(disableCls)s")}
        })''' % {'jsData': jsData, 'jqId': self.jqId, 'disableCls': self._report.style.cssName("CssLabelContainerDisabled"),
                 'breadCrumVar': self._report.jsGlobal.breadCrumVar, 'htmlCode': self.htmlId, 'reset': json.dumps(reset)}

  def jsAdd(self, jsData='data', jsDataKey=None, isPyData=False, jsParse=False, jsFnc=None, isUnique=False):
    """

    :example: c.jsAdd([{"value": 'A', "name": 'Test'}], isPyData=True, isUnique=True)
    """
    jsData = self._jsData(jsData, jsDataKey, jsParse, isPyData, jsFnc)
    return '''
      var existingCols = {};
      if (%(unique)s){%(jqId)s.find('span').each(function(){
        var pItem = $(this); 
        existingCols[pItem.data('content')] = true})};
      %(jsData)s.forEach(function(rec){
        if ((!%(unique)s) || (%(unique)s && (!(rec.value in existingCols))) ){
          var style = {'margin': 0, 'color': rec.color, 'display': 'block', 'position': 'relative', 'cursor': 'pointer'};
          var strCss = []; for (key in style) { strCss.push( key + ":" + style[key])};
          checkData = '&nbsp;';
          if (rec.checked){var checkData = '<i class="'+ %(jsStyles)s.icon + '" style="margin:2px"></i>'};
          var spanContent = '<span data-content="'+ rec.value + '" style="width:16px;display:inline-block;float:left;margin:0">'+ checkData +'</span><p style="margin:0" title="'+ rec.dsc + '">' + rec.name + '</p>';
          %(jqId)s.append($('<label style="' + strCss.join(";") + '">'+ spanContent +'</label>'))}
      })''' % {'jsData': jsData, 'jsStyles': json.dumps(self._jsStyles), 'jqId': self.jqId, 'unique': json.dumps(isUnique)}

  def jsRemove(self, jsData='data', jsDataKey=None, isPyData=False, jsParse=False, jsFnc=None):
    """

    :example: c.jsRemove([1, "A"], isPyData=True)
    """
    jsData = self._jsData(jsData, jsDataKey, jsParse, isPyData, jsFnc)
    return '''
      const ldata = %(jsData)s;
      if (ldata === true) {%(jqId)s.empty(); %(breadCrumVar)s['params']['%(htmlCode)s'] = []}
      else {
        %(jqId)s.find('span').each(function(){
          var pItem = $(this).next(); var bcIndex = %(breadCrumVar)s['params']['%(htmlCode)s'].indexOf(pItem.text());
          if(bcIndex > -1) {delete %(breadCrumVar)s['params']['%(htmlCode)s'].splice(bcIndex, 1)};
          if (ldata.indexOf($(this).data('content')) > -1){$(this).parent().remove()}
        })}''' % {'jsData': jsData, 'jqId': self.jqId, 'breadCrumVar': self._report.jsGlobal.breadCrumVar, 'htmlCode': self.htmlId}

  def jsItemState(self, jsData='data', jsDataKey=None, isPyData=False, jsParse=False, jsFnc=None):
    """

    :example: c.jsItemState(jsData=True, isPyData=True)
    :example: c.jsItemState(jsData=False, isPyData=True)
    """
    jsData = self._jsData(jsData, jsDataKey, jsParse, isPyData, jsFnc)
    return '''
      var ldata = %(jsData)s;
      %(jqId)s.find('span').each(function(){
        var itemCode = $(this).data('content');
        if(typeof ldata === "boolean"){
          if (ldata === true && $(this).find("i").attr("class") === undefined){$(this).trigger("click")}
          if (!ldata && $(this).find("i").attr("class") !== undefined){$(this).trigger("click")}}
        else if (ldata.indexOf(itemCode) > -1){if ($(this).find("i").attr("class") === undefined){$(this).trigger("click")}}
      }) ''' % {'jsData': jsData, 'jqId': self.jqId}

  def onDocumentLoadFnc(self):
    """ Pure Javascript onDocumentLoad Function """
    self.addGlobalFnc("%s(htmlObj, data, jsStyles)" % self.__class__.__name__, ''' htmlObj.empty() ;
      %(breadCrumVar)s['params'][htmlObj.attr('id')] = [];
      data.forEach(function(rec){
        if (rec.color == undefined) {rec.color = 'inherit'}
        var style = {'margin': 0, 'color': rec.color, 'display': 'block', 'position': 'relative', 'cursor': 'pointer'};
        if (rec.style != undefined) { for (key in rec.style) { style[key] = rec.style[key] }} ;
        if (rec.dsc == undefined) { rec.dsc = ''}
        if (rec.name == undefined) { rec.name = rec.value}
        var strCss = []; for (key in style) { strCss.push( key + ":" + style[key])};
        if (rec.checked == true) { 
          %(breadCrumVar)s['params'][htmlObj.attr('id')].push(rec.value );
          var spanContent = '<span data-content="' + rec.value + '" style="display:inline-block;float:left;margin:0"><i class="'+ jsStyles.icon + '" style="margin:2px"></i></span><p style="margin:0;padding:0" title="'+ rec.dsc + '">' + rec.name + '</p>'; }
        else {var spanContent = '<span data-content="'+ rec.value + '" style="width:16px;display:inline-block;float:left;margin:0">&nbsp;</span><p style="margin:0" title="'+ rec.dsc + '">' + rec.name + '</p>';}     
        htmlObj.append($('<label style="' + strCss.join(";") + '">'+ spanContent +'</label>'))}); htmlObj.find("p").tooltip(); 
        if (jsStyles.tooltip != ""){ 
          var tip = $('<i class="fas fa-info-circle" style="right:0" title="'+ jsStyles.tooltip +'"></i>') ;
          tip.tooltip(); htmlObj.append($("<div style='width:100%%;text-align:right'></div>").append(tip) )}
      ''' % {'breadCrumVar': self._report.jsGlobal.breadCrumVar}, 'Javascript Object builder')

  def jsEvents(self):
    if hasattr(self, 'jsFncFrag'):
      for eventKey, fnc in self.jsFncFrag.items():
        if self.htmlCode is not None:
          fnc.insert(0, self.jsAddUrlParam(self.htmlCode, self.val, isPyData=False))
        self._report.jsOnLoadEvtsFnc.add('''
          $( document ).on('%(eventKey)s', '#%(htmlId)s label', function(event) {
            if(($(this).find("i").attr("class") !== undefined) && ($(this).find("i").attr("class").includes('fas fa-times'))) {return {}};
            var useAsync = false; var isChecked = false; var htmlContent = $(this).find('span').find('i').length; 
            if (htmlContent == 0) {$(this).find('span').html('<i class="%(icon)s" style="padding:2px"></i>'); isChecked = true} else {$(this).find('span').html('<div style="width:16px;display:inline-block">&nbsp;</div>')}
            var data = %(data)s ; var returnVal = undefined;
            %(jsInfo)s; %(jsFnc)s; 
            if (!useAsync) {
              var body_loading_count = parseInt($('#body_loading span').text()); $('#body_loading span').html(body_loading_count - 1);
              if ($('#body_loading span').html() == '0') {$('#body_loading').remove()} }
            if (returnVal != undefined) {return returnVal}
          })''' % {'htmlId': self.htmlId, 'eventKey': eventKey, 'data': self.jsQueryData,
                   'jsFnc': ";".join([f for f in fnc if f is not None]), 'icon': self._jsStyles['icon'],
                   'jsInfo': self._report.jsInfo('process(es) running', 'body_loading')})

  def __str__(self):
    if self.selectAll:
      self.addGlobalFnc("CheckBoxSelectAll(event, srcObj, htmlId)", '''
        $(srcObj).find('div').toggleClass("fa-check");
        var ldata = $(srcObj).find('div').hasClass("fa-check");
        $('#'+ htmlId).find('span').each(function(){
          var itemCode = $(this).data('content');
          if(typeof ldata === "boolean"){
            if (ldata === true && $(this).find("i").attr("class") === undefined){$(this).trigger("click")}
            if (!ldata && $(this).find("i").attr("class") !== undefined){$(this).trigger("click")}}
          else if (ldata.indexOf(itemCode) > -1){if ($(this).find("i").attr("class") === undefined){$(this).trigger("click")}}
        })''')
      selectTag = '''
        <div onclick='event.stopPropagation();CheckBoxSelectAll(event, this, "%(htmlId)s");' style='vertical-align:middle;white-space:nowrap;margin-bottom:5px;cursor:pointer;padding:0'>
          <div style='box-sizing:border-box;margin:0;padding:0;border:1px solid %(color)s;font-size:10px;width:11px;height:11px;display:inline-block;color:%(color)s' class="fas">&nbsp;</div>
          <label style='cursor:inherit;margin:0;padding:0;font-style:italic;color:%(color)s;white-space:nowrap'>Select All</label>
        </div>''' % {"color": self._report.theme.colors[5], 'htmlId': self.htmlId}

    return '<div %(strAttr)s><div name="checks"></div></div>' % {'strAttr': self.get_attrs(pyClassNames=self.style.get_classes())}

  def to_word(self, document):
    from docx.shared import RGBColor

    selections = self._report.http.get(self.htmlCode).split(',')
    for rec in self.vals:
      p = document.add_paragraph(style='ListBullet')
      runner = p.add_run(rec['value'])
      if rec['value'] in selections:
        runner.font.color.rgb = RGBColor(0x42, 0x24, 0xE9)

  def to_xls(self, workbook, worksheet, cursor):
    selections = self._report.http.get(self.htmlCode).split(',')
    cellTitle = self.title if self.title != '' else 'Checkbox:'
    cell_format = workbook.add_format({'bold': True})
    worksheet.write(cursor['row'], 0, cellTitle, cell_format)
    cursor['row'] += 1
    for rec in self.vals:
      cell_format = workbook.add_format({})
      if rec['value'] in selections:
        cell_format = workbook.add_format({'bold': True, 'font_color': self._report.theme.colors[5]})
      worksheet.write(cursor['row'], 0, rec['value'], cell_format)
      cursor['row'] += 1
    cursor['row'] += 1

  @staticmethod
  def matchMarkDown(val):
    return re.match("\+\+\+(.*)", val) or re.match("\-\-\-(.*)", val)

  @classmethod
  def convertMarkDown(cls, val, regExpResult, report):
    param = True if val.startswith("+") else False
    if report is not None:
      getattr(report, cls.callFnc)(param, regExpResult.group(1))
    return ["report.%s(%s, %s)" % (cls.callFnc, param, regExpResult.group(1))]

  @classmethod
  def jsMarkDown(self, vals):
    if vals:
      return "+++" % self.label

    return "---%s" % self.label


class CheckButton(Html.Html):
  name, category, callFnc = 'Check Button', 'Button', 'check'

  def __init__(self, report, flag, tooltip, width, height, icon, label, htmlCode, options, profile):
    super(CheckButton, self).__init__(report, 'Y' if flag else 'N', htmlCode=htmlCode,
                                      css_attrs={"width": width, "height": height}, profile=profile)
    self.input = report.ui.images.icon("fas fa-check" if flag else "fas fa-times").css({"width": "12px"})
    if flag:
      self.input.style.css.color = self._report.theme.success[1]
    else:
      self.input.style.css.color = self._report.theme.danger[1]
    self.input.style.middle()
    self.input.inReport = False
    self.isDisable = options.get("disable", False)
    self.add_label(label, {"width": "none", "float": "none"}, position="after")
    self.add_icon(icon, {"float": 'none'}, position="after")
    self.css({'display': 'inline-block', 'margin-right': '10px'})
    if tooltip is not None:
      self.tooltip(tooltip)

  @property
  def dom(self):
    """
    Description:
    ------------
    The Javascript Dom object

    :rtype: JsHtml.JsHtmlButtonMenu
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtmlButtonMenu(self, report=self._report)
    return self._dom

  @property
  def js(self):
    """
    Description:
    -----------

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Autocomplete
    """
    if self._js is None:
      self._js = JsComponents.CheckButton(self, report=self._report)
    return self._js

  @property
  def _js__builder__(self):
    return ''' htmlObj.empty();
      if (data === true || data == 'Y'){
        htmlObj.append('<i class="fas fa-check" style="color:%(green)s;margin-bottom:2px;margin-left:2px"></i>');
        htmlObj.parent().data('isChecked', true)} 
      else {
        htmlObj.append('<i class="fas fa-times" style="font-size:14px;margin-top:2px;margin-left:5px;color:%(red)s"></i>');
        htmlObj.parent().data('isChecked', false)
      }''' % {'green': self._report.theme.success[1], 'red': self._report.theme.danger[1]}

  @property
  def style(self):
    """
    Description:
    ------------
    Property to the CSS Style of the component

    :rtype: GrpClsButton.ClassButtonCheckBox
    """
    if self._styleObj is None:
      self._styleObj = GrpClsButton.ClassButtonCheckBox(self)
    return self._styleObj

  def click(self, jsFncsTrue, jsFncFalse=None, withColors=True, profile=False):
    """
    Description:
    ------------
    Click even on the checkbox item

    Usage::

      ch = rptObj.ui.buttons.check(label="Label")
    ch.click(rptObj.js.alert("true"), rptObj.js.alert("false"))

    Attributes:
    ----------
    :param jsFncsTrue: Js function or a list of JsFunction to be triggered when checked
    :param jsFncFalse: Optional. Js function or a list of JsFunction to be triggered when unchecked

    :return: The htmlObl to allow the chaining
    """
    if self.label is not None and hasattr(self.label, 'style'):
      self.label.style.css.cursor = 'pointer'
    self.style.css.cursor = "pointer"
    if not isinstance(jsFncsTrue, list):
      jsFncsTrue = [jsFncsTrue]
    if jsFncFalse is None:
      jsFncFalse = []
    elif not isinstance(jsFncFalse, list):
      jsFncFalse = [jsFncFalse]
    if withColors:
      jsFncsTrue.append(self.input.dom.css({"color": self._report.theme.success[1]}).r)
      jsFncFalse.append(self.input.dom.css({"color": self._report.theme.danger[1]}).r)
    jsFncs = [
      self.input.dom.switchClass("fa-check", "fa-times"),
      JsIf.JsIf(self.input.dom.hasClass("fa-check"), jsFncsTrue).else_(
        jsFncFalse)]
    return super(CheckButton, self).click(jsFncs, profile)

  def __str__(self):
    return '''<div %s>%s</div>''' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.input.html())


class IconEdit(Html.Html):
  name, category, callFnc = 'Icon', 'Icons', 'iconEdit'
  cssCls = ["fa-layers", "fa-fw"]

  def __init__(self, report, position, icon, text, tooltip, width, height, htmlCode, profile):
    super(IconEdit, self).__init__(report, '', code=htmlCode, profile=profile,
                                   css_attrs={"width": width, 'height': height, 'float': 'left' if position is None else position})
    if tooltip is not None:
      self.tooltip(tooltip)
    # Add the internal components icons and helper
    self.add_span(text, css={"float": 'right'})
    self.add_icon(icon, {"color": self._report.theme.success[1], "margin": "2px", 'font-size': Defaults_css.font()})
    self.css({"margin": "5px 0", 'cursor': 'pointer'})

  def __str__(self):
    return "<span %s></span>" % (self.get_attrs(pyClassNames=self.style.get_classes()))


class Buttons(Html.Html):

  def __init__(self, report, data, color, width, height, htmlCode, helper, options, profile):
    super(Buttons, self).__init__(report, data, code=htmlCode, css_attrs={"width": width, "height": height, 'color': color},
                                  profile=profile)
    self.row = []
    for b in data:
      bt = report.ui.button(b, options={"group": "group_%s" % self.htmlId}).css({"margin-right": '5px'})
      bt.css(options.get("button_css", {}))
      bt.inReport = False
      self.row.append(bt)

  def __getitem__(self, i):
    """
    Description:
    ------------
    Get a button in this component

    Attributes:
    ----------
    :param i: Integer. The button index
    :rtype: Button
    """
    return self.row[i]

  def __str__(self):
    str_div = "".join([v.html() if hasattr(v, 'html') else v for v in self.row])
    return '<div %s>%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), str_div)


class ButtonMenuItem(object):

  def __init__(self, report, selector, parent):
    self._report, self._selector = report, selector
    self._src, self._js, self._events = parent, None, []

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript module of the items in the menu

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Autocomplete
    """
    if self._js is None:
      self._js = JsComponents.Menu(self._src, varName=self._selector, report=self._report)
    return self._js

  def on(self, event, jsFncs, profile=False):
    """
    Description:
    -----------
    Javascript generic events of the items in the menu

    Attributes:
    ----------
    :param event: String. The JavaScript event
    :param jsFncs: List: The Javascript fragments
    :param profile: Boolean.
    """
    self._events.append("%s.addEventListener('%s', function (event) { %s })" % (self._selector, event, JsUtils.jsConvertFncs(jsFncs, toStr=True)))
    return self._src

  def click(self, jsFncs):
    """
    Description:
    -----------
    Javascript click events of the items in the menu

    :param jsFncs: List: The Javascript fragments
    """
    self._events.append("%s.addEventListener('click', function (event) { %s })" % (self._selector, JsUtils.jsConvertFncs(jsFncs, toStr=True)))
    return self._src


class ButtonMenu(Html.Html):
  name, category, callFnc = 'ButtonMenu', 'buttons', 'menu'

  def __init__(self, report, record, text, icon, width, height, htmlCode, tooltip, profile, options):
    super(ButtonMenu, self).__init__(report, record, code=htmlCode, profile=profile, css_attrs={"width": width, "height": height})
    self.button = report.ui.button(text, icon, width, height, htmlCode, tooltip, profile, options)
    self.__items = {}
    self.button.inReport = False
    self.set_attrs(name="data-count", value=0)
    self.style.css.position = "relative"
    self.style.css.display = "inline-block"
    self.container = report.ui.div()
    self.container.inReport = False
    self.container.attr['class'].add("dropdown-content")
    self.container.style.css.display = "none"
    self.container.style.css.position = "absolute"
    self.container.style.css.z_index = 5
    self._jsStyles = {"padding": '5px', 'cursor': 'pointer', 'display': 'block'}

  def __getitem__(self, i):
    if i not in self.__items:
      self.__items[i] = ButtonMenuItem(self._report, "document.getElementById('%s').querySelectorAll('a')[%s]" % (self.htmlId, i), self)
    return self.__items[i]

  @property
  def _js__builder__(self):
    return '''
      var panel = htmlObj.querySelector("div");
      data.forEach(function(rec){
        var href = document.createElement("a"); href.innerHTML = rec;
        Object.keys(options).forEach(function(key){href.style[key] = options[key]});
        panel.appendChild(href)})'''

  @property
  def style(self):
    """
    Description:
    -----------
    Property to the CSS Style of the component

    Usage::

      self.style.css.margin = "5px"

    :rtype: GrpClsButton.ClassButton
    """
    if self._styleObj is None:
      self._styleObj = GrpClsButton.ClassButtonMenu(self)
    return self._styleObj

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    events = []
    for comp in self.__items.values():
      events.extend(comp._events)
    self.onReady(events)
    return '<div %s>%s%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.button.html(), self.container.html())



"""

"""

from epyk.core.Page import Report
from epyk.tests import data_urls
from epyk.core.css.themes import ThemeBlue
from epyk.core.css.themes import ThemeDark
from epyk.core.css.themes import ThemeRed
from epyk.core.css.themes import ThemeGreen
rptObj = Report()
# rptObj.ui.layouts.


#########THEMES#########
# rptObj.theme = ThemeBlue.BlueGrey()
# rptObj.theme = ThemeBlue.Blue()
# rptObj.theme = ThemeBlue.LightBlue()
# rptObj.theme = ThemeRed.Red()
# rptObj.theme = ThemeDark.Dark()
# rptObj.theme = ThemeDark.Grey()
# rptObj.theme = ThemeGreen.Green()
# rptObj.theme = ThemeGreen.LightGreen()
# rptObj.theme = ThemeGreen.Teal()

# f = rptObj.ui.forms.inputs([
#   {"label": "name", "htmlCode": "input"},
#   {"label": "name 2", "htmlCode": "input2"},
# ], "http://127.0.0.1:5000", "POST")
# f + rptObj.ui.fields.today('test')
# print(rptObj.tags.bdi("test"))
# print(rptObj.entities.POUND)
# print(rptObj.symbols.ALMOST_EQUAL_TO)
# rptObj.ui.d
# d = rptObj.ui.fields.today('test')
# i = rptObj.ui.fields.input(placeholder='test2', label='test1')
# i.input.set_attrs(name='required')
# i.input.set_attrs({'name': 'input1'})
# i2 = rptObj.ui.fields.input('test3', label='test2')
# rptObj.ui.fields.today('test')
# rptObj.ui.images.icon()
###########BUTTONS###########
# button = rptObj.ui.buttons.button('Contact Sales')
# imp = rptObj.ui.buttons.important('Get Started for Free')
# rptObj.ui.modal.forms([d, i, i2], "http://127.0.0.1:5000", "POST")
# rptObj.ui.modal.objects([d, i, i2], "POST", submit=False)
# disclaimer = rptObj.ui.texts.paragraph(disc)

# dis = rptObj.ui.modal.disclaimer(disc)
# button.click(rptObj.js.getElementById(dis.htmlId).css({'display': 'block'}))
# print(rptObj.js.addOnReady(rptObj.js.window.events.addClickListener(rptObj.js.if_('event.target == %s' % button.htmlId, button.), subEvents=['event'])))

#########FOOTER#########
# nav = rptObj.ui.navigation.bar(title="test")
# nav.add_text("This is a huge text that I don't know what to do with")
# nav + button
# nav + imp
# nav.add_text("Test text")
# nav + rptObj.ui.button("Click").css({"margin": '0 0 20px 5px'})
# rptObj.ui.navigation.banner("", "test", "google")
# footer = rptObj.ui.navigation.footer([button, imp])
#
# footer = rptObj.ui.navigation.complex_footer(4, {1: [button, imp], 2: [rptObj.ui.texts.text('toto')]})
# footer + button
# footer + imp
# footer + rptObj.ui.texts.paragraph('This is the end')
# import time
# start = time.time()

div = rptObj.ui.div("toto")
div.style.css_class.media({'.topnav li': {"float": None, 'width': '100%'},
                           'topnav li line': {'stroke-width': 0},
                           'topnav li [name=label]': {'width': '100%!IMPORTANT'}}, 'only', 'screen', {"and": [{'max-width': '600px'}]})
# div.style.css.background_color = 'yellow'
# div.style.css.height = '50px'

# div.style.css_class.media({div: {"background-color": "red"}}, 'only', 'screen', {"and": [{'max-width': '2000px'}]})
# print(div.style.css_class.get_ref())
#
# rptObj.js.addOnLoad()

# data_rest = rptObj.py.requests.csv(data_urls.COVID_ECDC_URL)
# print(data_rest)
# print(time.time() - start)


# rptObj.ui.texts.para

print(rptObj.outs.html_file(name='test'))





from epyk.core.css.styles.classes import CssStyle


class CssDrawer(CssStyle.Style):
  _selectors = {'child': 'div[name="drawer"]'}

  _attrs = {"position": "absolute", 'top': 0, 'height': '100%', 'overflow-x': 'hidden'}


class CssDrawerNav(CssStyle.Style):
  _selectors = {'child': 'div[name="drawer_nav"]'}

  _attrs = {"width": '100%', 'height': '30px', "background": 'blue', 'display': 'block'}


class CssDrawerHandle(CssStyle.Style):
  _selectors = {'child': 'div[name="drawer_handle"]'}

  _attrs = {"width": "15px", 'height': '100%', 'display': 'inline-block'}
  _after = {'position': 'absolute', 'top': '50%', 'padding': '2px', 'height': '100%'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.success[1]})


class CssDrawerContent(CssStyle.Style):
  _selectors = {'child': 'div[name="drawer_content"]'}

  _attrs = {'height': '100%', 'display': 'inline-block'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.success[0]})
